#!/usr/bin/env python
# -*- coding: latin-1 -*-

"""
"""

import pytest

import pydbc.db.model as model
from pydbc.parser import ParserWrapper
from pydbc.dbcListener import DbcListener


def check_dummy_node(node):
    assert node.rid == 0
    assert node.comment == 'Dummy node for non-existent senders/receivers.'
    assert node.name == 'Vector__XXX'
    assert node.node_id == 0

def test_no_nodes():
    parser = ParserWrapper('dbc', 'nodes', DbcListener, debug = False)
    DATA = "BU_:"
    session = parser.parseFromString(DATA)
    res = session.query(model.Node).all()
    assert len(res) == 1
    check_dummy_node(res[0])

def test_one_nodes():
    parser = ParserWrapper('dbc', 'nodes', DbcListener, debug = False)
    DATA = "BU_: MECU"
    session = parser.parseFromString(DATA)
    res = session.query(model.Node).all()
    assert len(res) == 2
    check_dummy_node(res[0])
    assert res[1].name == "MECU"

def test_multiple_nodes():
    parser = ParserWrapper('dbc', 'nodes', DbcListener, debug = False)
    DATA = "BU_: MECU KSG BCU"
    session = parser.parseFromString(DATA)
    res = session.query(model.Node).all()
    assert len(res) == 4
    check_dummy_node(res[0])
    assert res[1].name == "MECU"
    assert res[2].name == "KSG"
    assert res[3].name == "BCU"

def test_errorous_node1():
    parser = ParserWrapper('dbc', 'nodes', DbcListener, debug = False)
    DATA = "BU_: MECU 0235 BCU"
    session = parser.parseFromString(DATA)
    res = session.query(model.Node).all()
    assert len(res) == 2
    check_dummy_node(res[0])
    assert res[1].name == "MECU"

def test_errorous_node2():
    parser = ParserWrapper('dbc', 'nodes', DbcListener, debug = False)
    DATA = "BU_: MECU BCU 0235"
    session = parser.parseFromString(DATA)
    res = session.query(model.Node).all()
    assert len(res) == 3
    check_dummy_node(res[0])
    assert res[1].name == "MECU"
    assert res[2].name == "BCU"

def test_message1():
    parser = ParserWrapper('dbc', 'messages', DbcListener, debug = False)
    DATA = """BO_ 781 VBOX_10: 8 Vector__XXX
            SG_ Link_Time : 47|24@0+ (0.01,0) [0|0] "Seconds" Vector__XXX
            SG_ Status_remote : 39|8@0+ (1,0) [0|0] "-" Vector__XXX
            SG_ Sep_Angle : 7|32@0- (1,0) [-180|180] "Deg" Vector__XXX"""
    session = parser.parseFromString(DATA)
    res = session.query(model.Message).all()
    assert len(res) == 1
    msg = res[0]
    msg.comment == None
    msg.rid == 1
    msg.name == 'VBOX_10'
    msg.message_id == 781
    msg.dlc == 8
    msg.sender == 0
    msg.type == 'Message'
    signals = msg.signals
    assert len(signals) == 3
    s0, s1, s2 = signals
    assert s0.comment == None
    assert s0.rid == 1
    assert s0.name == 'Link_Time'
    assert s0.bitsize == 24
    assert s0.byteorder == 0
    assert s0.sign == 1
    assert s0.valuetype == 0
    assert s0.formula_factor == 0.01
    assert s0.formula_offset == 0.0
    assert s0.minimum == 0.0
    assert s0.maximum == 0.0
    assert s0.unit == 'Seconds'
    assert s1.comment == None
    assert s1.rid == 2
    assert s1.name == 'Status_remote'
    assert s1.bitsize == 8
    assert s1.byteorder == 0
    assert s1.sign == 1
    assert s1.valuetype == 0
    assert s1.formula_factor == 1.0
    assert s1.formula_offset == 0.0
    assert s1.minimum == 0.0
    assert s1.maximum == 0.0
    assert s1.unit == '-'
    assert s2.comment == None
    assert s2.rid == 3
    assert s2.name == 'Sep_Angle'
    assert s2.bitsize == 32
    assert s2.byteorder == 0
    assert s2.sign == -1
    assert s2.valuetype == 0
    assert s2.formula_factor == 1.0
    assert s2.formula_offset == 0.0
    assert s2.minimum == -180.0
    assert s2.maximum == 180.0
    assert s2.unit == 'Deg'

def test_message_no_signals():
    parser = ParserWrapper('dbc', 'messages', DbcListener, debug = False)
    DATA = """BO_ 781 VBOX_10: 8 Vector__XXX"""
    session = parser.parseFromString(DATA)
    res = session.query(model.Message).all()
    assert len(res) == 1
    assert res[0].signals == []


def test_malformed_message_no_id():
    parser = ParserWrapper('dbc', 'messages', DbcListener, debug = False)
    DATA = """BO_ VBOX_10: 8 Vector__XXX"""
    session = parser.parseFromString(DATA)
    res = session.query(model.Message).all()
    assert len(res) == 1
    msg = res[0]
    assert msg.comment == None
    assert msg.rid == 1
    assert msg.name == 'VBOX_10'
    assert msg.message_id == None
    assert msg.dlc == 8
    assert msg.sender == 0
    assert msg.type == 'Message'

def test_malformed_message_wrong_id():
    parser = ParserWrapper('dbc', 'messages', DbcListener, debug = False)
    DATA = """BO_ ABC VBOX_10: 8 Vector__XXX"""
    session = parser.parseFromString(DATA)
    res = session.query(model.Message).all()
    assert len(res) == 1
    msg = res[0]
    assert msg.comment == None
    assert msg.rid == 1
    assert msg.name == 'ABC'
    assert msg.message_id == None
    assert msg.dlc == 8
    assert msg.sender == 0
    assert msg.type == 'Message'

def test_malformed_message_no_name():
    parser = ParserWrapper('dbc', 'messages', DbcListener, debug = False)
    DATA = """BO_ 781 : 8 Vector__XXX"""
    session = parser.parseFromString(DATA)
    res = session.query(model.Message).all()
    assert len(res) == 1
    msg = res[0]
    assert msg.comment == None
    assert msg.rid == 1
    assert msg.name == None
    assert msg.message_id == 781
    assert msg.dlc == 8
    assert msg.sender == 0
    assert msg.type == 'Message'

def test_invalid_signal1():
    parser = ParserWrapper('dbc', 'messages', DbcListener, debug = False)
    DATA = """
    BO_ 1600 Reporting_Message_640: 8 Vector__XXX
        SG_ Opto_IO_1 : 0|1@0+ (1,0) [0|0] "" Vector__XXX
        SG_ Analog_Input_1_(0-26.4V) : 15|8@0+ (0.103529,0) [0|26] "" Vector__XXX
    """
    session = parser.parseFromString(DATA)
    res = session.query(model.Message).all()
    assert len(res) == 1
    msg = res[0]
    assert len(msg.signals) == 2
    s0, s1 = msg.signals
    assert s0.comment == None
    assert s0.rid == 1
    assert s0.name == 'Opto_IO_1'
    assert s0.bitsize == 1
    assert s0.byteorder == 0
    assert s0.sign == 1
    assert s0.valuetype == 0
    assert s0.formula_factor == 1.0
    assert s0.formula_offset == 0.0
    assert s0.minimum == 0.0
    assert s0.maximum == 0.0
    assert s0.unit == ''
    assert s1.comment == None
    assert s1.rid == 2
    assert s1.name == 'Analog_Input_1_'
    assert s1.bitsize == 0
    assert s1.byteorder == 0
    assert s1.sign == 1
    assert s1.valuetype == 0
    assert s1.formula_factor == 1.0
    assert s1.formula_offset == 0.0
    assert s1.minimum == 0.0
    assert s1.maximum == 0.0
    assert s1.unit == None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" """

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

    ( C) 2010-2020 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"

import io
import os
import pkgutil

import sqlalchemy as sa

from pydbc.logger import Logger
from pydbc.db import VNDB
from pydbc import parser
from pydbc.dbcListener import DbcListener

# from pydbc.ldfListener import LdfListener
# from pydbc.ncfListener import NcfListener
from pydbc.template import renderTemplateFromText
import pydbc.db.model as model


def fetch_attributes(db):
    # Attribute structure is currently to inconvenient for ad-hoc queries, so pre-fetch them.

    from collections import defaultdict
    from itertools import groupby
    from operator import itemgetter

    from pydbc.types import AttributeType, ValueType, CategoryType

    data = (
        db.session.query(
            model.Attribute_Value.num_value,
            model.Attribute_Value.string_value,
            model.Attribute_Definition.valuetype,
            model.Attribute_Definition.array,
            model.Attribute_Definition.objecttype,
            model.Attribute_Value.object_id,
            model.Attribute_Definition.name,
        )
        .join(model.Attribute_Definition)
        .order_by(model.Attribute_Definition.objecttype)
    )
    result = {}
    groups = []
    keyfunc = itemgetter(4)
    data = sorted(data, key=keyfunc)
    for k, g in groupby(data, keyfunc):
        group = list(g)
        result[AttributeType(k).name] = defaultdict(dict)
        groups.append(list(group))
        keyfunc = itemgetter(5)
        group2 = sorted(group, key=keyfunc)
        for k2, g2 in groupby(group2, keyfunc):
            for (
                num_value,
                string_value,
                value_type,
                array,
                object_type,
                object_id,
                name,
            ) in list(g2):
                if num_value is None and string_value is None:
                    value = None
                else:
                    if array:
                        conv = float if value_type == ValueType.FLOAT else int
                        value = [conv(x) for x in string_value.split(";")]
                    else:
                        if value_type in (ValueType.INT, ValueType.FLOAT):
                            conv = float if value_type == ValueType.FLOAT else int
                            value = conv(num_value)
                        else:
                            value = string_value
                item = {"value": value}
                result[AttributeType(k).name][k2][name] = value
    return result


class Exporter:
    """ """

    def __init__(self, name, encoding="latin-1", debug=False):
        self.logger = Logger(__name__)
        self.encoding = encoding
        if name == ":memory:":
            self.fnbase = "testfile.txt"
            self.db = VNDB(":memory:", debug=debug)
        else:
            self.fname = name.parts[-1]
            self.fnbase = name.stem
            self.fnext = name.suffix[1:].lower()
            self.fabsolute = name.absolute()

            self.db = VNDB(r"{}.vndb".format(self.fnbase), debug=debug)
        # res = renderTemplateFromText(self.TEMPLATE, namespace, formatExceptions = True, encoding = "utf-8" if ucout else "latin-1")

    def run(self):

        # xxx = self.db.session.query(model.Attribute_Value).join(model.Attribute_Definition).\
        #    filter(model.Attribute_Definition.name == "LIN_is_master").one()
        # node = self.db.session.query(model.Node).filter(model.Node.rid == xxx.object_id).one()
        # print("ATTRS:", fetch_attributes(self.db))
        namespace = dict(
            db=self.db, model=model, attributes=fetch_attributes(self.db), sa=sa
        )
        res = renderTemplateFromText(
            self.TEMPLATE, namespace, formatExceptions=False, encoding=self.encoding
        )
        # print("RES:", res)
        with io.open(
            "{}.render".format(self.fnbase), "w", encoding=self.encoding, newline="\r\n"
        ) as outf:
            outf.write(res)


class DbcExporter(Exporter):
    """ """

    TEMPLATE = pkgutil.get_data("pydbc", "cgen/templates/dbc.tmpl")


class LdfExporter(Exporter):
    """ """

    TEMPLATE = pkgutil.get_data("pydbc", "cgen/templates/ldf.tmpl")



# Generated from dbc.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .dbcParser import dbcParser
else:
    from dbcParser import dbcParser

# This class defines a complete listener for a parse tree produced by dbcParser.
class dbcListener(ParseTreeListener):

    # Enter a parse tree produced by dbcParser#dbcfile.
    def enterDbcfile(self, ctx:dbcParser.DbcfileContext):
        pass

    # Exit a parse tree produced by dbcParser#dbcfile.
    def exitDbcfile(self, ctx:dbcParser.DbcfileContext):
        pass


    # Enter a parse tree produced by dbcParser#messageTransmitters.
    def enterMessageTransmitters(self, ctx:dbcParser.MessageTransmittersContext):
        pass

    # Exit a parse tree produced by dbcParser#messageTransmitters.
    def exitMessageTransmitters(self, ctx:dbcParser.MessageTransmittersContext):
        pass


    # Enter a parse tree produced by dbcParser#messageTransmitter.
    def enterMessageTransmitter(self, ctx:dbcParser.MessageTransmitterContext):
        pass

    # Exit a parse tree produced by dbcParser#messageTransmitter.
    def exitMessageTransmitter(self, ctx:dbcParser.MessageTransmitterContext):
        pass


    # Enter a parse tree produced by dbcParser#signalExtendedValueTypeList.
    def enterSignalExtendedValueTypeList(self, ctx:dbcParser.SignalExtendedValueTypeListContext):
        pass

    # Exit a parse tree produced by dbcParser#signalExtendedValueTypeList.
    def exitSignalExtendedValueTypeList(self, ctx:dbcParser.SignalExtendedValueTypeListContext):
        pass


    # Enter a parse tree produced by dbcParser#signalExtendedValueType.
    def enterSignalExtendedValueType(self, ctx:dbcParser.SignalExtendedValueTypeContext):
        pass

    # Exit a parse tree produced by dbcParser#signalExtendedValueType.
    def exitSignalExtendedValueType(self, ctx:dbcParser.SignalExtendedValueTypeContext):
        pass


    # Enter a parse tree produced by dbcParser#messages.
    def enterMessages(self, ctx:dbcParser.MessagesContext):
        pass

    # Exit a parse tree produced by dbcParser#messages.
    def exitMessages(self, ctx:dbcParser.MessagesContext):
        pass


    # Enter a parse tree produced by dbcParser#message.
    def enterMessage(self, ctx:dbcParser.MessageContext):
        pass

    # Exit a parse tree produced by dbcParser#message.
    def exitMessage(self, ctx:dbcParser.MessageContext):
        pass


    # Enter a parse tree produced by dbcParser#signal.
    def enterSignal(self, ctx:dbcParser.SignalContext):
        pass

    # Exit a parse tree produced by dbcParser#signal.
    def exitSignal(self, ctx:dbcParser.SignalContext):
        pass


    # Enter a parse tree produced by dbcParser#receiver.
    def enterReceiver(self, ctx:dbcParser.ReceiverContext):
        pass

    # Exit a parse tree produced by dbcParser#receiver.
    def exitReceiver(self, ctx:dbcParser.ReceiverContext):
        pass


    # Enter a parse tree produced by dbcParser#transmitter.
    def enterTransmitter(self, ctx:dbcParser.TransmitterContext):
        pass

    # Exit a parse tree produced by dbcParser#transmitter.
    def exitTransmitter(self, ctx:dbcParser.TransmitterContext):
        pass


    # Enter a parse tree produced by dbcParser#multiplexerIndicator.
    def enterMultiplexerIndicator(self, ctx:dbcParser.MultiplexerIndicatorContext):
        pass

    # Exit a parse tree produced by dbcParser#multiplexerIndicator.
    def exitMultiplexerIndicator(self, ctx:dbcParser.MultiplexerIndicatorContext):
        pass


    # Enter a parse tree produced by dbcParser#valueTables.
    def enterValueTables(self, ctx:dbcParser.ValueTablesContext):
        pass

    # Exit a parse tree produced by dbcParser#valueTables.
    def exitValueTables(self, ctx:dbcParser.ValueTablesContext):
        pass


    # Enter a parse tree produced by dbcParser#valueTable.
    def enterValueTable(self, ctx:dbcParser.ValueTableContext):
        pass

    # Exit a parse tree produced by dbcParser#valueTable.
    def exitValueTable(self, ctx:dbcParser.ValueTableContext):
        pass


    # Enter a parse tree produced by dbcParser#valueDescription.
    def enterValueDescription(self, ctx:dbcParser.ValueDescriptionContext):
        pass

    # Exit a parse tree produced by dbcParser#valueDescription.
    def exitValueDescription(self, ctx:dbcParser.ValueDescriptionContext):
        pass


    # Enter a parse tree produced by dbcParser#nodes.
    def enterNodes(self, ctx:dbcParser.NodesContext):
        pass

    # Exit a parse tree produced by dbcParser#nodes.
    def exitNodes(self, ctx:dbcParser.NodesContext):
        pass


    # Enter a parse tree produced by dbcParser#bitTiming.
    def enterBitTiming(self, ctx:dbcParser.BitTimingContext):
        pass

    # Exit a parse tree produced by dbcParser#bitTiming.
    def exitBitTiming(self, ctx:dbcParser.BitTimingContext):
        pass


    # Enter a parse tree produced by dbcParser#newSymbols.
    def enterNewSymbols(self, ctx:dbcParser.NewSymbolsContext):
        pass

    # Exit a parse tree produced by dbcParser#newSymbols.
    def exitNewSymbols(self, ctx:dbcParser.NewSymbolsContext):
        pass


    # Enter a parse tree produced by dbcParser#version.
    def enterVersion(self, ctx:dbcParser.VersionContext):
        pass

    # Exit a parse tree produced by dbcParser#version.
    def exitVersion(self, ctx:dbcParser.VersionContext):
        pass


    # Enter a parse tree produced by dbcParser#objectValueTables.
    def enterObjectValueTables(self, ctx:dbcParser.ObjectValueTablesContext):
        pass

    # Exit a parse tree produced by dbcParser#objectValueTables.
    def exitObjectValueTables(self, ctx:dbcParser.ObjectValueTablesContext):
        pass


    # Enter a parse tree produced by dbcParser#objectValueTable.
    def enterObjectValueTable(self, ctx:dbcParser.ObjectValueTableContext):
        pass

    # Exit a parse tree produced by dbcParser#objectValueTable.
    def exitObjectValueTable(self, ctx:dbcParser.ObjectValueTableContext):
        pass


    # Enter a parse tree produced by dbcParser#environmentVariables.
    def enterEnvironmentVariables(self, ctx:dbcParser.EnvironmentVariablesContext):
        pass

    # Exit a parse tree produced by dbcParser#environmentVariables.
    def exitEnvironmentVariables(self, ctx:dbcParser.EnvironmentVariablesContext):
        pass


    # Enter a parse tree produced by dbcParser#environmentVariable.
    def enterEnvironmentVariable(self, ctx:dbcParser.EnvironmentVariableContext):
        pass

    # Exit a parse tree produced by dbcParser#environmentVariable.
    def exitEnvironmentVariable(self, ctx:dbcParser.EnvironmentVariableContext):
        pass


    # Enter a parse tree produced by dbcParser#accessNodes.
    def enterAccessNodes(self, ctx:dbcParser.AccessNodesContext):
        pass

    # Exit a parse tree produced by dbcParser#accessNodes.
    def exitAccessNodes(self, ctx:dbcParser.AccessNodesContext):
        pass


    # Enter a parse tree produced by dbcParser#environmentVariablesData.
    def enterEnvironmentVariablesData(self, ctx:dbcParser.EnvironmentVariablesDataContext):
        pass

    # Exit a parse tree produced by dbcParser#environmentVariablesData.
    def exitEnvironmentVariablesData(self, ctx:dbcParser.EnvironmentVariablesDataContext):
        pass


    # Enter a parse tree produced by dbcParser#environmentVariableData.
    def enterEnvironmentVariableData(self, ctx:dbcParser.EnvironmentVariableDataContext):
        pass

    # Exit a parse tree produced by dbcParser#environmentVariableData.
    def exitEnvironmentVariableData(self, ctx:dbcParser.EnvironmentVariableDataContext):
        pass


    # Enter a parse tree produced by dbcParser#signalTypes.
    def enterSignalTypes(self, ctx:dbcParser.SignalTypesContext):
        pass

    # Exit a parse tree produced by dbcParser#signalTypes.
    def exitSignalTypes(self, ctx:dbcParser.SignalTypesContext):
        pass


    # Enter a parse tree produced by dbcParser#signalType.
    def enterSignalType(self, ctx:dbcParser.SignalTypeContext):
        pass

    # Exit a parse tree produced by dbcParser#signalType.
    def exitSignalType(self, ctx:dbcParser.SignalTypeContext):
        pass


    # Enter a parse tree produced by dbcParser#comments.
    def enterComments(self, ctx:dbcParser.CommentsContext):
        pass

    # Exit a parse tree produced by dbcParser#comments.
    def exitComments(self, ctx:dbcParser.CommentsContext):
        pass


    # Enter a parse tree produced by dbcParser#comment.
    def enterComment(self, ctx:dbcParser.CommentContext):
        pass

    # Exit a parse tree produced by dbcParser#comment.
    def exitComment(self, ctx:dbcParser.CommentContext):
        pass


    # Enter a parse tree produced by dbcParser#attributeDefinitions.
    def enterAttributeDefinitions(self, ctx:dbcParser.AttributeDefinitionsContext):
        pass

    # Exit a parse tree produced by dbcParser#attributeDefinitions.
    def exitAttributeDefinitions(self, ctx:dbcParser.AttributeDefinitionsContext):
        pass


    # Enter a parse tree produced by dbcParser#attributeDefinition.
    def enterAttributeDefinition(self, ctx:dbcParser.AttributeDefinitionContext):
        pass

    # Exit a parse tree produced by dbcParser#attributeDefinition.
    def exitAttributeDefinition(self, ctx:dbcParser.AttributeDefinitionContext):
        pass


    # Enter a parse tree produced by dbcParser#relativeAttributeDefinitions.
    def enterRelativeAttributeDefinitions(self, ctx:dbcParser.RelativeAttributeDefinitionsContext):
        pass

    # Exit a parse tree produced by dbcParser#relativeAttributeDefinitions.
    def exitRelativeAttributeDefinitions(self, ctx:dbcParser.RelativeAttributeDefinitionsContext):
        pass


    # Enter a parse tree produced by dbcParser#relativeAttributeDefinition.
    def enterRelativeAttributeDefinition(self, ctx:dbcParser.RelativeAttributeDefinitionContext):
        pass

    # Exit a parse tree produced by dbcParser#relativeAttributeDefinition.
    def exitRelativeAttributeDefinition(self, ctx:dbcParser.RelativeAttributeDefinitionContext):
        pass


    # Enter a parse tree produced by dbcParser#attributeValueType.
    def enterAttributeValueType(self, ctx:dbcParser.AttributeValueTypeContext):
        pass

    # Exit a parse tree produced by dbcParser#attributeValueType.
    def exitAttributeValueType(self, ctx:dbcParser.AttributeValueTypeContext):
        pass


    # Enter a parse tree produced by dbcParser#attributeDefaults.
    def enterAttributeDefaults(self, ctx:dbcParser.AttributeDefaultsContext):
        pass

    # Exit a parse tree produced by dbcParser#attributeDefaults.
    def exitAttributeDefaults(self, ctx:dbcParser.AttributeDefaultsContext):
        pass


    # Enter a parse tree produced by dbcParser#attributeDefault.
    def enterAttributeDefault(self, ctx:dbcParser.AttributeDefaultContext):
        pass

    # Exit a parse tree produced by dbcParser#attributeDefault.
    def exitAttributeDefault(self, ctx:dbcParser.AttributeDefaultContext):
        pass


    # Enter a parse tree produced by dbcParser#relativeAttributeDefaults.
    def enterRelativeAttributeDefaults(self, ctx:dbcParser.RelativeAttributeDefaultsContext):
        pass

    # Exit a parse tree produced by dbcParser#relativeAttributeDefaults.
    def exitRelativeAttributeDefaults(self, ctx:dbcParser.RelativeAttributeDefaultsContext):
        pass


    # Enter a parse tree produced by dbcParser#relativeAttributeDefault.
    def enterRelativeAttributeDefault(self, ctx:dbcParser.RelativeAttributeDefaultContext):
        pass

    # Exit a parse tree produced by dbcParser#relativeAttributeDefault.
    def exitRelativeAttributeDefault(self, ctx:dbcParser.RelativeAttributeDefaultContext):
        pass


    # Enter a parse tree produced by dbcParser#attributeValue.
    def enterAttributeValue(self, ctx:dbcParser.AttributeValueContext):
        pass

    # Exit a parse tree produced by dbcParser#attributeValue.
    def exitAttributeValue(self, ctx:dbcParser.AttributeValueContext):
        pass


    # Enter a parse tree produced by dbcParser#attributeValues.
    def enterAttributeValues(self, ctx:dbcParser.AttributeValuesContext):
        pass

    # Exit a parse tree produced by dbcParser#attributeValues.
    def exitAttributeValues(self, ctx:dbcParser.AttributeValuesContext):
        pass


    # Enter a parse tree produced by dbcParser#attributeValueForObject.
    def enterAttributeValueForObject(self, ctx:dbcParser.AttributeValueForObjectContext):
        pass

    # Exit a parse tree produced by dbcParser#attributeValueForObject.
    def exitAttributeValueForObject(self, ctx:dbcParser.AttributeValueForObjectContext):
        pass


    # Enter a parse tree produced by dbcParser#relativeAttributeValues.
    def enterRelativeAttributeValues(self, ctx:dbcParser.RelativeAttributeValuesContext):
        pass

    # Exit a parse tree produced by dbcParser#relativeAttributeValues.
    def exitRelativeAttributeValues(self, ctx:dbcParser.RelativeAttributeValuesContext):
        pass


    # Enter a parse tree produced by dbcParser#relativeAttributeValueForObject.
    def enterRelativeAttributeValueForObject(self, ctx:dbcParser.RelativeAttributeValueForObjectContext):
        pass

    # Exit a parse tree produced by dbcParser#relativeAttributeValueForObject.
    def exitRelativeAttributeValueForObject(self, ctx:dbcParser.RelativeAttributeValueForObjectContext):
        pass


    # Enter a parse tree produced by dbcParser#signalGroups.
    def enterSignalGroups(self, ctx:dbcParser.SignalGroupsContext):
        pass

    # Exit a parse tree produced by dbcParser#signalGroups.
    def exitSignalGroups(self, ctx:dbcParser.SignalGroupsContext):
        pass


    # Enter a parse tree produced by dbcParser#signalGroup.
    def enterSignalGroup(self, ctx:dbcParser.SignalGroupContext):
        pass

    # Exit a parse tree produced by dbcParser#signalGroup.
    def exitSignalGroup(self, ctx:dbcParser.SignalGroupContext):
        pass


    # Enter a parse tree produced by dbcParser#categoryDefinitions.
    def enterCategoryDefinitions(self, ctx:dbcParser.CategoryDefinitionsContext):
        pass

    # Exit a parse tree produced by dbcParser#categoryDefinitions.
    def exitCategoryDefinitions(self, ctx:dbcParser.CategoryDefinitionsContext):
        pass


    # Enter a parse tree produced by dbcParser#categoryDefinition.
    def enterCategoryDefinition(self, ctx:dbcParser.CategoryDefinitionContext):
        pass

    # Exit a parse tree produced by dbcParser#categoryDefinition.
    def exitCategoryDefinition(self, ctx:dbcParser.CategoryDefinitionContext):
        pass


    # Enter a parse tree produced by dbcParser#categories.
    def enterCategories(self, ctx:dbcParser.CategoriesContext):
        pass

    # Exit a parse tree produced by dbcParser#categories.
    def exitCategories(self, ctx:dbcParser.CategoriesContext):
        pass


    # Enter a parse tree produced by dbcParser#category.
    def enterCategory(self, ctx:dbcParser.CategoryContext):
        pass

    # Exit a parse tree produced by dbcParser#category.
    def exitCategory(self, ctx:dbcParser.CategoryContext):
        pass


    # Enter a parse tree produced by dbcParser#intValue.
    def enterIntValue(self, ctx:dbcParser.IntValueContext):
        pass

    # Exit a parse tree produced by dbcParser#intValue.
    def exitIntValue(self, ctx:dbcParser.IntValueContext):
        pass


    # Enter a parse tree produced by dbcParser#floatValue.
    def enterFloatValue(self, ctx:dbcParser.FloatValueContext):
        pass

    # Exit a parse tree produced by dbcParser#floatValue.
    def exitFloatValue(self, ctx:dbcParser.FloatValueContext):
        pass


    # Enter a parse tree produced by dbcParser#number.
    def enterNumber(self, ctx:dbcParser.NumberContext):
        pass

    # Exit a parse tree produced by dbcParser#number.
    def exitNumber(self, ctx:dbcParser.NumberContext):
        pass


    # Enter a parse tree produced by dbcParser#stringValue.
    def enterStringValue(self, ctx:dbcParser.StringValueContext):
        pass

    # Exit a parse tree produced by dbcParser#stringValue.
    def exitStringValue(self, ctx:dbcParser.StringValueContext):
        pass


    # Enter a parse tree produced by dbcParser#identifierValue.
    def enterIdentifierValue(self, ctx:dbcParser.IdentifierValueContext):
        pass

    # Exit a parse tree produced by dbcParser#identifierValue.
    def exitIdentifierValue(self, ctx:dbcParser.IdentifierValueContext):
        pass



del dbcParser


#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__  = 'Christoph Schueler'
__version__ = '0.1.0'


import re

from sqlalchemy.sql.expression import literal, bindparam
from sqlalchemy.ext import baked

from pydbc.logger import Logger
from pydbc import parser
from pydbc.types import AttributeType, BusType, CategoryType, ValueType
from pydbc.db.model import (
    Dbc_Version, Message, Message_Signal, Network, Node, Signal, Value_Description,
    Valuetable, EnvironmentVariablesData, EnvVar, Attribute_Definition, Attribute_Value,
    Node_TxMessage, Node_RxSignal, Category_Definition, Category_Value, AttributeRel_Value,
    Signal_Group_Signal, Signal_Group, Vndb_Protocol, Object_Valuetable, VndbMeta
)


## Attributes
##  GenSigStartValue    ==> to set initial value until first message is received.
##  GenMsgCycleTime     ==> to set transmission rate

"""
Before moving on, note that J1939 is a bit special in regards to the CAN DBC file format:
- The ID is extended 29 bit - in DBC context, this means that the the leading bit of the ID is "flipped" and needs to be re-flipped
- Secondly, the relevant ID is the PGN, i.e. a sub part of the CAN ID (start bit 9, length 18)
"""

DIGITS = re.compile(r'(\d+)')

CO_MPX = re.compile(r"^(?P<multiplexed>m)?(?P<value>\d+)?(?P<multiplexer>M)?$")


def extractAccessType(value):
    match = DIGITS.search(value)
    if match:
        return int(match.group())
    else:
        return None


class DbcListener(parser.BaseListener):
    """

    """

    def __init__(self, database, logLevel = 'INFO', *args, **kws):
        super(DbcListener, self).__init__(database, logLevel, *args, **kws)
        self.logger = Logger(__name__, level = logLevel)
        self.bake_queries()
        self.insertNetwork()

    def bake_queries(self):
        self.MESSAGE_BY_MESSAGE_ID = self.bakery(lambda session: session.query(Message).filter(Message.message_id == bindparam('message_id')))
        self.ENVVAR_BY_NAME = self.bakery(lambda session: session.query(EnvVar).filter(EnvVar.name == bindparam('name')))
        self.ENVDATA_BY_NAME = self.bakery(lambda session: session.query(EnvironmentVariablesData).filter(
            EnvironmentVariablesData.name == bindparam('name')))
        self.MESSAGE_SIGNAL_BY_NAME = self.bakery(lambda session: session.query(Signal.rid).join(Message_Signal).join(Message).\
                filter(Message.message_id == bindparam('messageID'), Signal.name == bindparam('signal_name')))
        self.NETWORK_BY_RID = self.bakery(lambda session: session.query(Network).filter(Network.rid == bindparam('rid')))
        self.SIGNAL_BY_RID = self.bakery(lambda session: session.query(Signal).filter(Signal.rid == bindparam('rid')))
        self.EXISTS_ATTRIBUTE_DEFINITION = self.bakery(lambda session: session.query(literal(True)).filter(
            Attribute_Definition.name == bindparam('name')))
        self.EXISTS_ATTRIBUTE_VALUE = self.bakery(lambda session: session.query(literal(True)).filter(
                Attribute_Value.object_id == bindparam('rid'), Attribute_Value.attribute_definition_id == bindparam('ad')))
        self.EXISTS_NETWORK = self.bakery(lambda session: session.query(literal(True)).filter(Network.name == bindparam('name')))
        self.EXISTS_NODE_RXSIGNAL = self.bakery(lambda session: session.query(literal(True)).filter(
                Node_RxSignal.node_id == bindparam('nodeId'),
                Node_RxSignal.message_id == bindparam('messageId'),
                Node_RxSignal.signal_id == bindparam('signalId')))
        self.EXISTS_VALUE_DESCRIPTION = self.bakery(lambda session: session.query(literal(True)).filter(
                Value_Description.valuetable_id == bindparam('rid'), Value_Description.value == bindparam('value')))
        self.MESSAGE_SIGNAL_BY_NAME2 = self.bakery(lambda session: session.query(Signal).join(Message_Signal).join(Message).\
            filter(Message.message_id == bindparam('messageID'), Signal.name == bindparam('signalName')))

    def getAttributeType(self, value):
        ATS = {
            "NETWORK": AttributeType.NETWORK,
            "BU": AttributeType.NODE,
            "BO": AttributeType.MESSAGE,
            "SG": AttributeType.SIGNAL,
            "EV": AttributeType.ENV_VAR,
            "REL_NODE": AttributeType.REL_NODE,
            "REL_SIGNAL": AttributeType.REL_SIGNAL,
            "REL_ENV_VAR": AttributeType.REL_ENV_VAR,
        }
        return ATS.get(value)

    def get_signal_by_name(self, messageID, signal_name):
        res = self.MESSAGE_SIGNAL_BY_NAME(self.session).params(messageID = messageID,signal_name = signal_name).first()
        if res:
            return res.rid
        else:
            self.logger.error("Message signal '{}'::'{}' does not exist.".format(messageID, signal_name))
            return []

    def insertReceivers(self, messageId, signalId, receiver):
        for rcv in receiver:
            res = self.NODE_BY_NAME(self.session).params(name = rcv).first()
            if res:
                nodeId = res.rid
            else:
                self.logger.error("While inserting signal receivers: node '{}' does not exist.".format(rcv))
                continue
            exists = self.EXISTS_NODE_RXSIGNAL(self.session).params(nodeId = nodeId, messageId = messageId,
                signalId = signalId).first()
            if exists:
                self.logger.error("While inserting signal receivers: receiver messageId: {} signalId: {} nodeId: {} already exists.".
                    format(messageId, signalId, nodeId)
                )
            else:
                rxs = Node_RxSignal(node_id = nodeId, message_id = messageId, signal_id = signalId)
                self.session.add(rxs)
        self.session.flush()

    def insertValueDescription(self, rid, description):
        objs = []
        for desc, value in description:
            exists = self.EXISTS_VALUE_DESCRIPTION(self.session).params(rid = rid, value = value).first()
            if exists:
                pass
            else:
                vd = Value_Description(valuetable_id = rid, value = value, value_description = desc)
                objs.append(vd)
        self.session.add_all(objs)
        self.session.flush()

    def insertAttributeDefinitions(self, ctx):
        self.log_insertion("AttributeDefinitions")
        ctx.value = [x.value for x in ctx.items]
        for attr in ctx.value:
            attrType = attr['type']
            name = attr['name']
            vt = attr['value']['type']
            values = attr['value']['value']
            minimum = None
            maximum = None
            enumvalues = None
            defaultNumber = None
            defaultString = None
            if attrType == 'BU_':
                objType = AttributeType.NODE
            elif attrType == 'BO_':
                objType = AttributeType.MESSAGE
            elif attrType == 'SG_':
                objType = AttributeType.SIGNAL
            elif attrType == 'EV_':
                objType = AttributeType.ENV_VAR
            elif attrType is None:
                objType = AttributeType.NETWORK
            elif attrType == 'BU_BO_REL_':
                objType = AttributeType.REL_NODE
            elif attrType == 'BU_SG_REL_':
                objType = AttributeType.REL_SIGNAL
            elif attrType == 'BU_EV_REL_':
                objType = AttributeType.REL_ENV_VAR
            if vt == 'INT':
                valueType = ValueType.INT
                minimum, maximum = values
            elif vt == 'HEX':
                valueType = ValueType.HEX
                minimum, maximum = values
            elif vt == 'FLOAT':
                valueType = ValueType.FLOAT
                minimum, maximum = values
            elif vt == 'STRING':
                valueType = ValueType.STRING
            elif vt == 'ENUM':
                valueType = ValueType.ENUM
                enumvalues = ';'.join(values)
            exists = self.EXISTS_ATTRIBUTE_DEFINITION(self.session).params(name = name).first()
            if exists:
                self.logger.error("While inserting attribute definitions: an attribute definition named '{}' already exists.".format(name))
            else:
                ad = Attribute_Definition(name = name, objecttype = objType, valuetype = valueType, minimum = minimum,
                                          maximum = maximum, enumvalues = enumvalues)
                self.session.add(ad)
        self.session.flush()

    def insertAttributeDefaults(self, ctx):
        self.log_insertion("AttributeDefaults")
        defaults = {}
        for item in ctx.items:
            name, value = item.value
            defaults[name] = value
            ad = self.ATTRIBUTE_DEFINITION_BY_NAME(self.session).params(name = name).first()
            if not ad:
                self.logger.error("While inserting attribute default values: attribute '{}' does not exist.".format(name))
                continue
            if ad.valuetype in (ValueType.INT, ValueType.HEX, ValueType.FLOAT):
                ad.default_number = value
            elif ad.valuetype in (ValueType.STRING, ValueType.ENUM):
                ad.default_string = value
        self.session.flush()
        ctx.value = defaults

    def insertNetwork(self, specific = None):
        self.log_insertion("Network")
        name = self.db.dbname
        exists = self.EXISTS_NETWORK(self.session).params(name = name).first()
        if exists:
            self.logger.error("While inserting network: an network named '{}' already exists.".format(name))
            return
        network = Network(name = name)
        self.session.add(network)
        proto = Vndb_Protocol(network = network, name = BusType.CAN.name, specific = specific)
        self.session.add(proto)
        self.session.flush()
        self.network_id = network.rid

    def exitDbcfile(self, ctx):
        self.session.commit()
        self.value = dict(
            version=ctx.version().value,
            newSymbols=ctx.newSymbols().value,
            bitTiming=ctx.bitTiming().value,
            nodes=ctx.nodes().value,
            valueTables=ctx.valueTables().value,
            messages=ctx.messages().value,
            messageTransmitters=ctx.messageTransmitters().value,
            environmentVariables=ctx.environmentVariables().value,
            environmentVariablesData=ctx.environmentVariablesData().value,
            signalTypes=ctx.signalTypes().value,
            comments=ctx.comments().value,
            attributeDefinitions=ctx.attributeDefinitions().value,
            relativeAttributeDefinitions=ctx.relativeAttributeDefinitions().value,
            attributeDefaults=ctx.attributeDefaults().value,
            relativeAttributeDefaults=ctx.relativeAttributeDefaults().value,
            attributeValues=ctx.attributeValues().value,
            relativeAttributeValues=ctx.relativeAttributeValues().value,
            objectValueTables=ctx.objectValueTables().value,
            categoryDefinitions=ctx.categoryDefinitions().value,
            categories=ctx.categories().value,
            signalGroups=ctx.signalGroups().value,
            signalExtendedValueTypeList=ctx.signalExtendedValueTypeList().value

        )
        meta_entry = VndbMeta(schema_version=1, vndb_type=1)
        self.session.add(meta_entry)
        self.session.commit()
        self.logger.info("Finished processing.")
        #self.logger.info(
        #    f"Finished parsing. Network '{self.network.name}' with {len(self.network.messages)} messages."
        #)

    def exitMessageTransmitters(self, ctx):
        self.log_insertion("NodeTxMessages")
        ctx.value = [x.value for x in ctx.items]
        for transmitter in ctx.value:
            msg = self.MESSAGE_BY_MESSAGE_ID(self.session).params(message_id = transmitter['messageID']).first()
            if not msg:
                self.logger.error("While inserting message transmitters: message '{}' does not exist.".format(transmitter['messageID']))
                continue
            for name in transmitter['transmitters']:
                node = self.NODE_BY_NAME(self.session).params(name = name).first()
                if not node:
                    self.logger.error("While inserting message transmitters: node '{}' does not exist.".format(name))
                    continue
                ntm = Node_TxMessage(node = node, message = msg)
                self.session.add(ntm)
        self.session.flush()

    def exitMessageTransmitter(self, ctx):
        transmitters = self.getValue(ctx.tx)
        ctx.value = dict(messageID = self.getValue(ctx.messageID), transmitters = transmitters)

    def exitSignalExtendedValueTypeList(self, ctx):
        self.log_insertion("SignalExtendedValueTypes")
        ctx.value = [x.value for x in ctx.items]
        valueTypes = ctx.value
        for item in valueTypes:
            msgId = item['messageID']
            sigName = item['signalName']
            vt = item['valueType']
            srid = self.get_signal_by_name(msgId, sigName)
            signal = self.SIGNAL_BY_RID(self.session).params(rid = srid).first()
            if not signal:
                self.logger.error("While inserting signal extended value types: signal '{}' does not exist.".format(srid))
                continue
            signal.valuetype = vt
        self.session.flush()

    def exitSignalExtendedValueType(self, ctx):
        messageID = self.getValue(ctx.messageID)
        signalName = self.getValue(ctx.signalName)
        valType = self.getValue(ctx.valType)
        if not valType in (0, 1, 2, 3):
            self.logger.error("While parsing signal extended value type: value type must be in range [0..3] - got '{}', using 0.".format(valType))
            valType = 0
        ctx.value = dict(messageID = messageID, signalName = signalName, valueType = valType)

    def exitMessages(self, ctx):
        self.log_insertion("Messages and Signals")
        ctx.value = [x.value for x in ctx.items]
        for msg in ctx.value:
            name = msg['name']
            mid = msg['messageID']
            dlc = msg['dlc']
            signals = msg['signals']
            transmitter = msg['transmitter']
            tx_node = self.NODE_BY_NAME(self.session).params(name = transmitter).first()
            if not tx_node:
                self.logger.error("While inserting messages: node '{}' does not exist.".format(transmitter))
                continue
            mm = Message(name = name, message_id = mid, dlc = dlc, sender = tx_node.rid)
            self.session.add(mm)
            for signal in signals:
                name = signal['name']
                startBit = signal['startBit']
                signalSize = signal['signalSize']
                byteOrder = signal['byteOrder']
                sign = signal['sign']
                factor = signal['factor']
                offset = signal['offset']
                minimum = signal['minimum']
                maximum = signal['maximum']
                unit = signal['unit']
                receiver = signal['receiver']
                multiplexorSignal = None
                multiplexDependent = None
                multiplexorValue = None
                multiplexerIndicator = signal['multiplexerIndicator']
                if multiplexerIndicator:
                    match = CO_MPX.match(multiplexerIndicator)
                    if match:
                        gd = match.groupdict()
                        #print("MPX:", gd)
                        gd['multiplexed'] == 'm'
                        multiplexorValue = gd['value']
                        multiplexorSignal = 1 if gd['multiplexer'] == 'M' else 0
                        multiplexDependent = 0 if multiplexorSignal else 1

                    """
                    multiplexorSignal = 1 if multiplexerIndicator == 'M' else 0
                    if multiplexorSignal:
                        multiplexDependent = 0
                        multiplexorValue = None
                    else:
                        multiplexDependent = 1

                        try:
                            multiplexorValue = int(multiplexerIndicator[1 : ])
                        except Exception as e:
                            print(str(e), "MPX:", multiplexerIndicator)
                else:
                    multiplexorSignal = None
                    multiplexDependent = None
                    multiplexorValue = None
                     """

                ss = Signal(name = name, bitsize = signalSize, byteorder = byteOrder, sign = sign,
                    formula_factor = factor, formula_offset = offset, minimum = minimum, maximum = maximum, unit = unit
                )
                self.session.add(ss)
                srid = ss.rid
                ms = Message_Signal(offset = startBit, multiplexor_signal = multiplexorSignal,
                    multiplex_dependent = multiplexDependent, multiplexor_value = multiplexorValue,
                    signal = ss, message = mm)
                self.session.add(ms)
                self.session.flush()
                self.insertReceivers(mm.rid, ss.rid, receiver)

    def exitMessage(self, ctx):
        ctx.value = dict(messageID = self.getValue(ctx.messageID), name = self.getValue(ctx.messageName),
            dlc = self.getValue(ctx.messageSize), transmitter = self.getTerminal(ctx.transmt),
            signals = [x.value for x in ctx.sgs]
        )

    def exitSignal(self, ctx):
        byteOrder = self.getValue(ctx.byteOrder)
        if not byteOrder in (0, 1):
            self.logger.error("While parsing signal: Error while parsing signal '{}': byteorder must be either 0 or 1 -- using 0".format(ctx.signalName.value))
            byteOrder = 0
        name = self.getValue(ctx.signalName)
        startBit = self.getValue(ctx.startBit)
        signalSize = self.getValue(ctx.signalSize)
        st = self.getTerminal(ctx.sign)
        sign = None if st is None else -1 if st == '-' else +1
        factor = self.getValue(ctx.factor)
        offset = self.getValue(ctx.offset)
        minimum = self.getValue(ctx.minimum)
        maximum = self.getValue(ctx.maximum)
        unit = self.getValue(ctx.unit)
        receiver = self.getValue(ctx.rcv) or []
        ctx.value = dict(name = name, startBit = startBit, signalSize = signalSize,
            byteOrder = byteOrder, sign = sign, factor = factor, offset = offset,
            minimum = minimum, maximum = maximum, unit = unit, receiver = receiver,
            multiplexerIndicator = ctx.mind.value if ctx.mind else None
        )

    def exitReceiver(self, ctx):
        ctx.value = [ctx.fid.text] + [x.value for x in ctx.ids]

    def exitTransmitter(self, ctx):
        ctx.value = [x.value for x in ctx.ids]

    def exitMultiplexerIndicator(self, ctx):
        ctx.value = self.getValue(ctx.mind)

    def exitValueTables(self, ctx):
        self.log_insertion("ValueTables")
        ctx.value = [x.value for x in ctx.items]
        for table in ctx.value:
            name = table['name']
            description = table['description']
            vt = Valuetable(name = name)
            self.session.add(vt)
            self.session.flush()
            self.insertValueDescription(vt.rid, description)
        self.session.flush()

    def exitValueTable(self, ctx):
        ctx.value = dict(name = ctx.name.value, description = [x.value for x in ctx.desc])

    def exitValueDescription(self, ctx):
        ctx.value = (ctx.name.value, ctx.val.value, )

    def exitNodes(self, ctx):
        self.log_insertion("Nodes")
        ctx.value = [x.value for x in ctx.ids]
        nodeSet = set(['Vector__XXX'])
        for name in ctx.value:
            if not name in nodeSet:
                nn = Node(name = name)
                self.session.add(nn)
            nodeSet.add(name)
        self.session.flush()

    def exitBitTiming(self, ctx):
        ctx.value = dict(baudrate = self.getValue(ctx.baudrate), btr1 = self.getValue(ctx.btr1), btr2 = self.getValue(ctx.btr2))

    def exitNewSymbols(self, ctx):
        ctx.value = [x.text for x in ctx.ids]

    def exitVersion(self, ctx):
        self.log_insertion("Versions")
        ctx.value = self.getValue(ctx.vs)
        version = ctx.value
        network = self.network_id
        vers = Dbc_Version(version_string = version, network = network)
        self.session.add(vers)
        self.session.flush()

    def exitObjectValueTables(self, ctx):
        self.log_insertion("ObjectValuetables")
        ctx.value = [x.value for x in ctx.items]
        for table in ctx.value:
            tp = table['type']
            description = table['description']
            if tp == 'SG':
                object_rid = table['messageID']
                name = table['signalName']
                otype = 0
            elif tp == 'EV':
                name = table['envVarName']
                otype = 1
                env_var = self.ENVVAR_BY_NAME(self.session).params(name = name).first()
                if not env_var:
                    self.logger.error("While inserting object value tables: environment variable '{}' does not exist.".format(name))
                    continue
                object_rid = env_var.rid
            vt = Valuetable(name = name)
            self.session.add(vt)
            self.session.flush()
            ovt = Object_Valuetable(object_type= otype, object_rid = object_rid, valuetable_id = vt.rid)
            self.session.add(ovt)
            self.session.flush()
            self.insertValueDescription(vt.rid, description)

    def exitObjectValueTable(self, ctx):
        items = [x.value for x in ctx.items]
        if ctx.messageID:
            messageID = ctx.messageID.value
            signalName = ctx.signalName.value
            di = dict(messageID = messageID, signalName = signalName)
            tp = "SG"
        else:
            envVarName = ctx.envVarName.value
            di = dict(envVarName = envVarName)
            tp = "EV"
        ctx.value = dict(type = tp, description = items, **di)

    def exitEnvironmentVariables(self, ctx):
        self.log_insertion("EnvVars")
        # TODO: Process after EVDATA!!!
        ctx.value = [x.value for x in ctx.evs]
        for var in ctx.value:
            unit = var['unit']
            initialValue = var['initialValue']
            accessNodes = var['accessNodes']
            accessType = var['accessType']
            minimum = var['minimum']
            maximum = var['maximum']
            envId = var['envId']
            varType = var['varType']
            name = var['name']
            env_data = self.ENVDATA_BY_NAME(self.session).params(name = name).first()
            if not env_data:
                #self.logger.error("Environment variable data '{}' does not exist.".format(name))
                #continue
                dataSize = 0
            else:
                dataSize = env_data.value
            envVar = EnvVar(name = name, type = varType, unit = unit, minimum = minimum, maximum = maximum,
                access = accessType, startup_value = initialValue, size = dataSize)
            self.session.add(envVar)
            for node in accessNodes:
                nn = self.NODE_BY_NAME(self.session).params(name = node).first()
                if not nn:
                    self.logger.error("While inserting environment variables: node '{}' does not exist.".format(name))
                    continue
                envVar.accessingNodes.append(nn)
        self.session.flush()

    def exitEnvironmentVariable(self, ctx):
        accessType = extractAccessType(ctx.DUMMY_NODE_VECTOR().getText())

        ctx.value = dict(name = self.getValue(ctx.name), varType = self.getValue(ctx.varType),
            minimum = self.getValue(ctx.minimum), maximum = self.getValue(ctx.maximum), unit = self.getValue(ctx.unit),
            initialValue = self.getValue(ctx.initialValue), envId = self.getValue(ctx.envId),
            accessType = accessType, accessNodes = self.getValue(ctx.accNodes)
        )

    def exitAccessNodes(self, ctx):
        nodes = [x.value for x in ctx.items]
        ctx.value = nodes

    def exitEnvironmentVariablesData(self, ctx):
        self.log_insertion("EnvironmentVariablesData")
        ctx.value = [x.value for x in ctx.evars]
        for item in ctx.value:
            print("EVD", item)
            name = item['name']
            value = item['value']
            evd = EnvironmentVariablesData(name = name, value = value)
            self.session.add(evd)
        self.session.flush()

    def exitEnvironmentVariableData(self, ctx):
        ctx.value = dict(name = self.getValue(ctx.varname), value = self.getValue(ctx.value))

    def exitSignalTypes(self, ctx):
        self.log_insertion("SignalTypes")
        ctx.value =[x.value for x in ctx.sigTypes]

    def exitSignalType(self, ctx):
        ctx.value = dict(name = self.getValue(ctx.signalTypeName), size = self.getValue(ctx.signalSize),
            byteOrder = self.getValue(ctx.byteOrder), valueType = self.getValue(ctx.valueType),
            factor = self.getValue(ctx.factor), offset = self.getValue(ctx.offset), minimum = self.getValue(ctx.minimum),
            maximum = self.getValue(ctx.maximum), unit = self.getValue(ctx.unit), defaultValue = self.getValue(ctx.defaultValue),
            valTable = self.getValue(ctx.valTable),
        )

    def exitComments(self, ctx):
        self.log_insertion("Comments")
        ctx.value = [x.value for x in ctx.items]
        for comment in ctx.value:
            tp = comment['type']
            text = comment['comment']
            key = comment['key']
            if tp == 'BU':
                obj = self.NODE_BY_NAME(self.session).params(name = key).first()
                if not obj:
                    self.logger.error("While inserting comments: node '{}' does not exist.".format(key))
                    continue
                obj.comment = text
            elif tp == 'BO':
                obj = self.MESSAGE_BY_MESSAGE_ID(self.session).params(message_id = key).first()
                if not obj:
                    self.logger.error("While inserting comments: message '{}' does not exist.".format(key))
                    continue
                obj.comment = text
            elif tp == 'SG':
                rid = self.get_signal_by_name(*key)
                if not rid:
                    self.logger.error("While inserting comments: message signal '{}' does not exist.".format(key))
                    continue
                obj = self.session.query(Signal).filter(Signal.rid == rid).first()
                obj.comment = text
            elif tp == 'EV':
                obj = self.ENVVAR_BY_NAME(self.session).params(name = key).first()
                if not obj:
                    self.logger.error("While inserting comments: environment variable '{}' does not exist.".format(key))
                    continue
                obj.comment = text
            else:   # NW !?
                obj = self.NETWORK_BY_RID(self.session).params(rid = self.network_id).first()
                if not obj:
                    self.logger.error("While inserting comments: network '{}' does not exist.".format(self.network_id))
                    continue
                obj.comment = text
        self.session.flush()

    def exitComment(self, ctx):
        comment = self.getValue(ctx.s)
        if ctx.c0:
            tp = "BU"
            key = self.getValue(ctx.c0)
        elif ctx.i1:
            tp = "BO"
            key = self.getValue(ctx.i1)
        elif ctx.i2:
            tp = "SG"
            key = (self.getValue(ctx.i2), self.getValue(ctx.c2), )
        elif ctx.c3:
            tp = "EV"
            key = self.getValue(ctx.c3)
        else:
            tp = "NETWORK"
            key = None
        ctx.value = dict(type = tp, key = key, comment = comment)

    def exitAttributeDefinitions(self, ctx):
        self.insertAttributeDefinitions(ctx)

    def exitAttributeDefinition(self, ctx):
        objectType = self.getTerminal(ctx.objectType)
        attributeName = self.getValue(ctx.attrName)
        attributeValue = self.getValue(ctx.attrValue)
        ctx.value = dict(type = objectType, name = attributeName, value = attributeValue)

    def exitRelativeAttributeDefinitions(self, ctx):
        ctx.value = [x.value for x in ctx.items]
        self.insertAttributeDefinitions(ctx)

    def exitRelativeAttributeDefinition(self, ctx):
        objectType = self.getTerminal(ctx.objectType)
        attributeName = self.getValue(ctx.attrName)
        attributeValue = self.getValue(ctx.attrValue)
        ctx.value = dict(type = objectType, name = attributeName, value = attributeValue)

    def exitAttributeValueType(self, ctx):
        if ctx.i00:
            tp = "INT"
            value = (self.getValue(ctx.i00), self.getValue(ctx.i01), )
        elif ctx.i10:
            tp = "HEX"
            value = (self.getValue(ctx.i10), self.getValue(ctx.i11), )
        elif ctx.f0:
            tp = "FLOAT"
            value = (float(self.getValue(ctx.f0)), float(self.getValue(ctx.f1)), )
        elif ctx.s0:
            tp = "STRING"
            value = None
        elif ctx.efirst:
            tp = "ENUM"
            efirst = [self.getValue(ctx.efirst)]
            eitems = [x.value for x in ctx.eitems]
            value = efirst + eitems
        ctx.value = dict(type = tp, value = value)

    def exitAttributeDefaults(self, ctx):
        self.insertAttributeDefaults(ctx)

    def exitAttributeDefault(self, ctx):
        name = self.getValue(ctx.n)
        value = self.getValue(ctx.v)
        ctx.value = (name, value)

    def exitRelativeAttributeDefaults(self, ctx):
        self.insertAttributeDefaults(ctx)

    def exitRelativeAttributeDefault(self, ctx):
        name = self.getValue(ctx.n)
        value = self.getValue(ctx.v)
        ctx.value = (name, value)

    def exitAttributeValue(self, ctx):
        if ctx.s:
            ctx.value = self.getValue(ctx.s)
        elif ctx.n:
            ctx.value = self.getValue(ctx.n)

    def exitAttributeValues(self, ctx):
        self.log_insertion("AttributeValues")
        ctx.value = [x.value for x in ctx.items]
        values = []
        for attr in ctx.value:
            stringValue = None
            numValue = None
            value = attr['attributeValue']
            if isinstance(value, str):
                stringValue = value
            else:
                numValue = value
            aname = attr['name']
            ad = self.ATTRIBUTE_DEFINITION_BY_NAME(self.session).params(name = attr['name']).first()
            if not ad:
                self.logger.error("While inserting attribute values: attribute definition named '{}' does not exist.".format(attr['name']))
                continue
            attrType = self.getAttributeType(attr['attributeType'])
            if attrType == AttributeType.MESSAGE:
                key = attr['messageID']
                msg = self.MESSAGE_BY_MESSAGE_ID(self.session).params(message_id = key).first()
                if not msg:
                    self.logger.error("While inserting attribute values: message '{}' does not exist.".format(key))
                    continue
                rid = msg.rid
            elif attrType == AttributeType.SIGNAL:
                key = attr['messageID'], attr['signalName'],
                rid = self.get_signal_by_name(*key)
            elif attrType == AttributeType.NODE:
                key = attr['nodeName']
                node = self.NODE_BY_NAME(self.session).params(name = key).first()
                if not node:
                    self.logger.error("While inserting attribute values: node '{}' does not exist.".format(key))
                    continue
                rid = node.rid
            elif attrType == AttributeType.ENV_VAR:
                key = attr['envVarname']
                envvar = self.ENVVAR_BY_NAME(self.session).params(name = key).first()
                if not envvar:
                    self.logger.error("While inserting attribute values: environment variable '{}' does not exist.".format(key))
                    continue
                rid = envvar.rid
            elif attrType == AttributeType.NETWORK:
                key = ''
                rid = 0
            else:
                rid = 0
                key = ''
            exists = self.EXISTS_ATTRIBUTE_VALUE(self.session).params(rid = rid, ad = ad.rid).first()
            if exists:
                self.logger.error("While inserting attribute values: attribute value for {} {} {} already exists.".format(AttributeType.SIGNAL.name, key, ad.name))
            else:
                av = Attribute_Value(object_id = rid, attribute_definition = ad, num_value = numValue, string_value = stringValue)
                values.append(av)
        self.session.add_all(values)
        self.session.flush()

    def exitAttributeValueForObject(self, ctx):
        attributeName = self.getValue(ctx.attributeName)
        if ctx.nodeName:
            nodeName = self.getValue(ctx.nodeName)
            attrValue = self.getValue(ctx.buValue)
            di = dict(attributeType = 'BU', nodeName = nodeName)
        elif ctx.mid1:
            mid1 = self.getValue(ctx.mid1)
            attrValue = self.getValue(ctx.boValue)
            di = dict(attributeType = 'BO', messageID = mid1)
        elif ctx.mid2:
            mid2 = self.getValue(ctx.mid2)
            signalName = self.getValue(ctx.signalName)
            attrValue = self.getValue(ctx.sgValue)
            di = dict(attributeType = 'SG', messageID = mid2, signalName = signalName)
        elif ctx.evName:
            evName = self.getValue(ctx.evName)
            attrValue = self.getValue(ctx.evValue)
            di = dict(attributeType = 'EV', envVarname = evName)
        else:
            evName = None
            attrValue = self.getValue(ctx.attrValue)
            di = dict(attributeType = "NETWORK")
        ctx.value = dict(name = attributeName, attributeValue = attrValue, **di)

    def exitRelativeAttributeValues(self, ctx):
        self.log_insertion("RelativeAttributeValues")
        items = [x.value for x in ctx.items]
        ctx.value = items
        for attr in items:
            attributeName = attr['attributeName']
            attributeValue = attr['attributeValue']
            attrributeType = self.getAttributeType(attr['attributeType'])
            ad = self.ATTRIBUTE_DEFINITION_BY_NAME(self.session).params(name = attr['attributeName']).first()
            if not ad:
                self.logger.error("While inserting relative attribute values: attribute definition '{}' does not exist.".format(attr['attributeName']))
                continue
            node = self.NODE_BY_NAME(self.session).params(name = attr['nodeName']).first()
            if not ad:
                self.logger.error("While inserting relative attribute values: node '{}' does not exist.".format(attr['nodeName']))
                continue
            nodeId = node.rid
            parent = attr['parent']
            optOid1 = nodeId
            optOid2 = None
            stringValue = None
            numValue = None
            value = attr['attributeValue']
            if isinstance(value, str):
                stringValue = value
            else:
                numValue = value
            if attrributeType == AttributeType.REL_SIGNAL:
                messageID = parent['messageID']
                optOid2 = messageID
                rid = self.get_signal_by_name(messageID, parent['signalName'])
                if not rid:
                    self.logger.error("While inserting relative attribute values: message signal '{}'::'{}' does not exist".format(messageID, parent['signalName']))
                    continue
            elif attrributeType == AttributeType.REL_ENV_VAR:
                evName = parent['evName']
                env_var = self.ENVVAR_BY_NAME(self.session).params(name = evName).first()
                if not env_var:
                    self.logger.error("While inserting relative attribute values: environment variable '{}' does not exist.".format(evName))
                    continue
                rid = env_var.rid
                #optOid2 = ???
            elif attrributeType == AttributeType.REL_NODE:
                messageID = parent['messageID']
                optOid2 = messageID
                msg = self.MESSAGE_BY_MESSAGE_ID(self.session).params(message_id = messageID).first()
                if not msg:
                    self.logger.error("While inserting relative attribute values: message '{}' does not exist.".format(messageID))
                    continue
                rid = msg.rid
            arv = AttributeRel_Value(object_id = rid, attribute_definition = ad, num_value = numValue,
                string_value = stringValue, opt_object_id_1 = optOid1, opt_object_id_2 = optOid2
            )
            self.session.add(arv)
        self.session.flush()

    def exitRelativeAttributeValueForObject(self, ctx):
        attrType = self.getTerminal(ctx.attrType)
        attributeName = self.getValue(ctx.attributeName)
        nodeName = self.getValue(ctx.nodeName)
        if attrType == "BU_BO_REL_":
            messageID = self.getValue(ctx.nodeAddress)
            attributeType = "REL_NODE"
            parent = dict(messageID = messageID)
            attrValue = self.getValue(ctx.attrValue)
        elif attrType == "BU_SG_REL_":
            messageID = self.getValue(ctx.messageID)
            signalName = self.getValue(ctx.signalName)
            attributeType = "REL_SIGNAL"
            parent = dict(messageID = messageID, signalName = signalName)
            attrValue = self.getValue(ctx.attrValue)
        elif attrType == "BU_EV_REL_":
            evName = ctx.evName.value
            attributeType = "REL_ENV_VAR"
            parent = dict(evName = evName)
            attrValue = self.getValue(ctx.evValue)
        ctx.value = dict(
            attributeType = attributeType, attributeName = attributeName, attributeValue = attrValue,
            nodeName = nodeName, parent = parent
        )

    def exitSignalGroups(self, ctx):
        self.log_insertion("SignalGroups")
        items = [x.value for x in ctx.items]
        ctx.value = items
        for group in ctx.value:
            messageID = group['messageID']
            gValue = group['gvalue']
            signalNames = group['signals']
            groupName = group['groupName']
            msg = self.MESSAGE_BY_MESSAGE_ID(self.session).params(message_id = messageID).first()
            if not msg:
                self.logger.error("While inserting signal groups: message '{}' does not exist.".format(messageID))
                continue
            sg = Signal_Group(name = groupName, value = gValue, message = msg)
            self.session.add(sg)
            self.session.flush()
            for signalName in signalNames:
                signal = self.MESSAGE_SIGNAL_BY_NAME2(self.session).params(messageID = messageID, signalName = signalName).first()
                if not signal:
                    self.logger.error("While inserting signal groups: message signal '{}'::'{}' does not exist.".format(messageID, signalName))
                    continue
                sgs = Signal_Group_Signal(signal_group = sg, message = msg, signal = signal)
                self.session.add(sgs)
        self.session.flush()

    def exitSignalGroup(self, ctx):
        messageID = self.getValue(ctx.messageID)
        groupName = self.getValue(ctx.groupName)
        gvalue = self.getValue(ctx.gvalue)
        signals = [x.value for x in ctx.signals]
        ctx.value = dict(messageID = messageID, groupName = groupName, gvalue = gvalue, signals = signals)

    def exitCategoryDefinitions(self, ctx):
        self.log_insertion("CategoryDefinitions")
        ctx.value = [x.value for x in ctx.items]
        for category in ctx.value:
            print("CAT-DEF", category)
            cd = Category_Definition(name = category['name'], key = category['category'], level = category['value'])
            self.session.add(cd)
        self.session.flush()

    def exitCategoryDefinition(self, ctx):
        ctx.value = dict(name = self.getValue(ctx.name), category = self.getValue(ctx.cat), value = self.getValue(ctx.num))

    def exitCategories(self, ctx):
        self.log_insertion("Categories")
        ctx.value = [x.value for x in ctx.items]
        for category in ctx.value:
            print("CAT-VALUE", category)
            attrType = category['type']
            catId = category['category']
            if attrType == 'BU':
                nodeName = category['nodeName']
                objType = CategoryType.NODE
                node = self.NODE_BY_NAME(self.session).params(node = nodeName).first()
                if not node:
                    self.logger.error("While inserting categories: node '{}' does not exist.".format(nodeName))
                    continue
                rid = node.rid
            elif attrType == 'BO':
                objType = CategoryType.MESSAGE
                messageID = category['messageID']
                msg = self.MESSAGE_BY_MESSAGE_ID(self.session).params(message_id = messageID).first()
                if not msg:
                    self.logger.error("While inserting categories: message '{}' does not exist.".format(messageID))
                    continue
                rid = msg.rid
            elif attrType == 'EV':
                envVarname = category['envVarname']
                objType = CategoryType.ENV_VAR
                env_var = self.ENVVAR_BY_NAME(self.session).params(name = envVarname)
                if not env_var:
                    self.logger.error("While inserting categories: environment variable '{}' does not exist.".format(envVarname))
                    continue
                rid = env_var.rid
            cv = Category_Value(object_id = rid, category_definition_id = catId, objecttype = objType)
            self.session.add(cv)
            print(cv)
        self.session.flush()

    def exitCategory(self, ctx):
        category = self.getValue(ctx.cat)
        if ctx.nodeName:
            nodeName = self.getValue(ctx.nodeName)
            di = dict(type = 'BU', nodeName = nodeName)
        elif ctx.mid1:
            mid1 = self.getValue(ctx.mid1)
            di = dict(type = 'BO', messageID = mid1)
        elif ctx.evName:
            evName = self.getValue(ctx.evName)
            di = dict(type = 'EV', envVarname = evName)
        ctx.value = dict(category = category, **di)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code generators for standalone applications using Mako templates.

This module provides two generators:
- MicroPython CAN application (Python script) using machine.CAN or pyb.CAN depending on platform.
- Linux SocketCAN C source file with simple encode/decode and send/recv demo.

Both generators work from an SQLAlchemy session populated by pydbc (either
created via DBCCreator or returned by ParserWrapper/DbcListener). No runtime
DB access is required by the generated code; all message/signal metadata is
embedded at generation time.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Iterable, Tuple
import pkgutil

from sqlalchemy.orm import Session

from pydbc.db.model import (
    Message,
    Message_Signal,
    Signal,
    LinNetwork,
    LinMasterNode,
    LinSlaveNode,
    LinSignal,
    LinUnconditionalFrame,
    LinSignalRepresentation,
    LinSignalEncodingEntry_Physical,
    LinSignalEncodingEntry_Logical,
    LinScheduleTable,
    LinScheduleTable_Command_Frame,
    LinScheduleTable_Command_MasterReq,
    LinScheduleTable_Command_SlaveResp,
    LinScheduleTable_Command_AssignNad,
    LinScheduleTable_Command_ConditionalChangeNad,
    LinScheduleTable_Command_DataDump,
    LinScheduleTable_Command_SaveConfiguration,
    LinScheduleTable_Command_AssignFrameIdRange,
    LinScheduleTable_Command_FreeFormat,
    LinScheduleTable_Command_AssignFrameId,
    LinSporadicFrame,
    LinEventTriggeredFrame,
    LinConfigurableFrame,
)
from pydbc.template import renderTemplateFromText


@dataclass
class CGSignal:
    name: str
    start_bit: int
    size: int
    little_endian: bool
    signed: bool
    factor: float
    offset: float
    minimum: float
    maximum: float


@dataclass
class CGMessage:
    name: Optional[str]
    message_id: int
    dlc: int
    signals: List[CGSignal]


@dataclass
class CGLinSignalEncodingPhysical:
    min_value: Optional[float]
    max_value: Optional[float]
    scale: float
    offset: float
    text_info: Optional[str]


@dataclass
class CGLinSignalEncodingLogical:
    signal_value: float
    text_info: Optional[str]


@dataclass
class CGLinSignal:
    name: str
    size: int
    start_bit: int
    signed: bool
    init_value: int
    factor: float
    offset: float
    minimum: Optional[float]
    maximum: Optional[float]
    publisher: Optional[str]
    subscribers: List[str]
    physical_entries: List[CGLinSignalEncodingPhysical]
    logical_entries: List[CGLinSignalEncodingLogical]


@dataclass
class CGLinFrame:
    name: str
    frame_id: int
    pid: int
    dlc: int
    publisher: Optional[str]
    signals: List[CGLinSignal]


@dataclass
class CGLinScheduleCommand:
    command_type: str
    frame_time: float
    frame_name: Optional[str]
    node_name: Optional[str]
    params: Dict[str, Any]


@dataclass
class CGLinScheduleTable:
    name: str
    commands: List[CGLinScheduleCommand]


@dataclass
class CGLinSporadicFrame:
    name: str
    frame_id: int
    pid: int
    associated_frames: List[str]


@dataclass
class CGLinEventTriggeredFrame:
    name: str
    frame_id: int
    pid: int
    associated_frames: List[str]
    collision_resolving_schedule: Optional[str]


@dataclass
class CGLinConfigurableFrame:
    node_name: str
    frame_name: str
    identifier: Optional[int]


@dataclass
class CGLinNode:
    name: str
    role: str
    configured_nad: Optional[int]
    initial_nad: Optional[int]


def _collect_messages(
    session: Session, only: Optional[Iterable[str]] = None
) -> List[CGMessage]:
    q = session.query(Message)
    if only:
        q = q.filter(Message.name.in_(list(only)))
    msgs: List[CGMessage] = []
    for m in q.all():
        sigs: List[CGSignal] = []
        for ms in m.message_signals:
            s: Signal = ms.signal
            sigs.append(
                CGSignal(
                    name=s.name,
                    start_bit=ms.offset,
                    size=s.bitsize,
                    little_endian=True if (s.byteorder == 1) else False,
                    signed=(s.sign == 0),  # 0=signed, 1=unsigned in current model
                    factor=s.formula_factor or 1.0,
                    offset=s.formula_offset or 0.0,
                    minimum=float(s.minimum) if s.minimum is not None else 0.0,
                    maximum=float(s.maximum) if s.maximum is not None else 0.0,
                )
            )
        msgs.append(
            CGMessage(
                name=m.name,
                message_id=m.message_id or 0,
                dlc=m.dlc,
                signals=sigs,
            )
        )
    # Sort for stable output
    msgs.sort(key=lambda x: (x.message_id, x.name or ""))
    for msg in msgs:
        msg.signals.sort(key=lambda s: s.start_bit)
    return msgs


def _compute_lin_pid(frame_id: int) -> int:
    fid = frame_id & 0x3F
    p0 = ((fid >> 0) ^ (fid >> 1) ^ (fid >> 2) ^ (fid >> 4)) & 0x1
    p1 = (~((fid >> 1) ^ (fid >> 3) ^ (fid >> 4) ^ (fid >> 5))) & 0x1
    return fid | (p0 << 6) | (p1 << 7)


def _normalize_lin_init_value(value: Any) -> int:
    if isinstance(value, (list, tuple)):
        return int(value[0]) if value else 0
    return int(value)


def _collect_lin_signal_encodings(
    session: Session, signal: LinSignal
) -> Tuple[List[CGLinSignalEncodingPhysical], List[CGLinSignalEncodingLogical]]:
    physical_entries: List[CGLinSignalEncodingPhysical] = []
    logical_entries: List[CGLinSignalEncodingLogical] = []
    reps = session.query(LinSignalRepresentation).filter_by(signal=signal).all()
    for rep in reps:
        entries = rep.signal_encoding_type.entries if rep.signal_encoding_type else []
        for entry in entries or []:
            if isinstance(entry, LinSignalEncodingEntry_Physical):
                physical_entries.append(
                    CGLinSignalEncodingPhysical(
                        min_value=entry.min_value,
                        max_value=entry.max_value,
                        scale=entry.scale if entry.scale is not None else 1.0,
                        offset=entry.offset if entry.offset is not None else 0.0,
                        text_info=entry.text_info,
                    )
                )
            elif isinstance(entry, LinSignalEncodingEntry_Logical):
                logical_entries.append(
                    CGLinSignalEncodingLogical(
                        signal_value=entry.signal_value,
                        text_info=entry.text_info,
                    )
                )
    return physical_entries, logical_entries


def _collect_lin_data(session: Session) -> Dict[str, Any]:
    signal_cache: Dict[int, CGLinSignal] = {}

    def build_signal(signal: LinSignal, start_bit: int) -> CGLinSignal:
        cached = signal_cache.get(signal.lin_signal_id)
        if cached and cached.start_bit == start_bit:
            return cached
        physical_entries, logical_entries = _collect_lin_signal_encodings(
            session, signal
        )
        if physical_entries:
            primary = physical_entries[0]
            factor = primary.scale
            offset = primary.offset
            minimum = primary.min_value
            maximum = primary.max_value
        else:
            factor = 1.0
            offset = 0.0
            minimum = None
            maximum = None
        cg = CGLinSignal(
            name=signal.name,
            size=signal.bitsize,
            start_bit=start_bit,
            signed=(signal.sign == 0),
            init_value=_normalize_lin_init_value(signal.init_value),
            factor=factor,
            offset=offset,
            minimum=minimum,
            maximum=maximum,
            publisher=signal.publisher.name if signal.publisher else None,
            subscribers=[s.name for s in signal.subscribers],
            physical_entries=physical_entries,
            logical_entries=logical_entries,
        )
        signal_cache[signal.lin_signal_id] = cg
        return cg

    unconditional_frames: List[CGLinFrame] = []
    for frame in session.query(LinUnconditionalFrame).all():
        if frame.__class__ is not LinUnconditionalFrame:
            continue
        sigs: List[CGLinSignal] = []
        for fs in frame.frame_signals:
            sigs.append(build_signal(fs.signal, fs.signal_offset))
        unconditional_frames.append(
            CGLinFrame(
                name=frame.name,
                frame_id=frame.frame_id,
                pid=_compute_lin_pid(frame.frame_id),
                dlc=frame.dlc,
                publisher=frame.publisher.name if frame.publisher else None,
                signals=sigs,
            )
        )

    sporadic_frames: List[CGLinSporadicFrame] = []
    for frame in session.query(LinSporadicFrame).all():
        sporadic_frames.append(
            CGLinSporadicFrame(
                name=frame.name,
                frame_id=frame.frame_id,
                pid=_compute_lin_pid(frame.frame_id),
                associated_frames=[f.name for f in frame.associated_frames],
            )
        )

    event_frames: List[CGLinEventTriggeredFrame] = []
    for frame in session.query(LinEventTriggeredFrame).all():
        event_frames.append(
            CGLinEventTriggeredFrame(
                name=frame.name,
                frame_id=frame.frame_id,
                pid=_compute_lin_pid(frame.frame_id),
                associated_frames=[f.name for f in frame.associated_frames],
                collision_resolving_schedule=(
                    frame.collision_resolving_schedule_table.name
                    if frame.collision_resolving_schedule_table
                    else None
                ),
            )
        )

    configurable_frames: List[CGLinConfigurableFrame] = []
    for cfg in session.query(LinConfigurableFrame).all():
        configurable_frames.append(
            CGLinConfigurableFrame(
                node_name=cfg.node.name if cfg.node else "",
                frame_name=cfg.frame.name if cfg.frame else "",
                identifier=cfg.identifier,
            )
        )

    schedule_tables: List[CGLinScheduleTable] = []
    for table in session.query(LinScheduleTable).all():
        commands: List[CGLinScheduleCommand] = []
        for cmd in table.commands:
            command_type = cmd.__class__.__name__.replace(
                "LinScheduleTable_Command_", ""
            )
            frame_name = None
            node_name = None
            params: Dict[str, Any] = {}
            if isinstance(cmd, LinScheduleTable_Command_Frame):
                frame_name = cmd.frame.name if cmd.frame else None
            elif isinstance(cmd, LinScheduleTable_Command_AssignNad):
                node_name = cmd.node.name if cmd.node else None
            elif isinstance(cmd, LinScheduleTable_Command_ConditionalChangeNad):
                params = {
                    "nad": cmd.nad,
                    "id": cmd.id,
                    "byte": cmd.byte,
                    "mask": cmd.mask,
                    "inv": cmd.inv,
                    "new_nad": cmd.new_nad,
                }
            elif isinstance(cmd, LinScheduleTable_Command_DataDump):
                node_name = cmd.node.name if cmd.node else None
                params = {
                    "d1": cmd.d1,
                    "d2": cmd.d2,
                    "d3": cmd.d3,
                    "d4": cmd.d4,
                    "d5": cmd.d5,
                }
            elif isinstance(cmd, LinScheduleTable_Command_SaveConfiguration):
                node_name = cmd.node.name if cmd.node else None
            elif isinstance(cmd, LinScheduleTable_Command_AssignFrameIdRange):
                node_name = cmd.node.name if cmd.node else None
                params = {
                    "frame_index": cmd.frame_index,
                    "frame_pid1": cmd.frame_pid1,
                    "frame_pid2": cmd.frame_pid2,
                    "frame_pid3": cmd.frame_pid3,
                    "frame_pid4": cmd.frame_pid4,
                }
            elif isinstance(cmd, LinScheduleTable_Command_FreeFormat):
                params = {
                    "d1": cmd.d1,
                    "d2": cmd.d2,
                    "d3": cmd.d3,
                    "d4": cmd.d4,
                    "d5": cmd.d5,
                    "d6": cmd.d6,
                    "d7": cmd.d7,
                    "d8": cmd.d8,
                }
            elif isinstance(cmd, LinScheduleTable_Command_AssignFrameId):
                node_name = cmd.node.name if cmd.node else None
                frame_name = cmd.frame.name if cmd.frame else None
            elif isinstance(
                cmd,
                (
                    LinScheduleTable_Command_MasterReq,
                    LinScheduleTable_Command_SlaveResp,
                ),
            ):
                pass

            commands.append(
                CGLinScheduleCommand(
                    command_type=command_type,
                    frame_time=cmd.frame_time,
                    frame_name=frame_name,
                    node_name=node_name,
                    params=params,
                )
            )
        schedule_tables.append(CGLinScheduleTable(name=table.name, commands=commands))

    nodes: List[CGLinNode] = []
    master = session.query(LinMasterNode).first()
    if master:
        nodes.append(
            CGLinNode(
                name=master.name,
                role="master",
                configured_nad=None,
                initial_nad=None,
            )
        )
    for slave in session.query(LinSlaveNode).all():
        nodes.append(
            CGLinNode(
                name=slave.name,
                role="slave",
                configured_nad=slave.configured_NAD,
                initial_nad=slave.initial_NAD,
            )
        )

    network = session.query(LinNetwork).first()
    return {
        "network_name": network.name if network else None,
        "nodes": nodes,
        "unconditional_frames": unconditional_frames,
        "sporadic_frames": sporadic_frames,
        "event_triggered_frames": event_frames,
        "configurable_frames": configurable_frames,
        "schedule_tables": schedule_tables,
    }


class MicroPythonCanAppGenerator:
    """Generates a standalone MicroPython CAN app (.py) with encode/decode helpers.

    Usage:
        gen = MicroPythonCanAppGenerator(session)
        code = gen.render(only=["EngineData"])  # optional filter
        Path("micropython_app.py").write_text(code, encoding="utf-8")
    """

    TEMPLATE = pkgutil.get_data("pydbc", "cgen/templates/micropython_can.py.tmpl")

    def __init__(self, session: Session):
        self._session = session

    def render(
        self, only: Optional[Iterable[str]] = None, app_name: str = "pydbc_mpy_app"
    ) -> str:
        msgs = _collect_messages(self._session, only)
        ns: Dict[str, Any] = {
            "app_name": app_name,
            "messages": msgs,
        }
        text = (
            self.TEMPLATE.decode("utf-8")
            if isinstance(self.TEMPLATE, (bytes, bytearray))
            else str(self.TEMPLATE)
        )
        return renderTemplateFromText(text, ns, encoding="utf-8")


class SocketCanCGenerator:
    """Generates a standalone Linux SocketCAN C program (.c) with encode/decode helpers.

    Usage:
        gen = SocketCanCGenerator(session)
        code = gen.render(only=["EngineData"])  # optional filter
        Path("socketcan_app.c").write_text(code, encoding="utf-8")
    """

    TEMPLATE = pkgutil.get_data("pydbc", "cgen/templates/socketcan.c.tmpl")

    def __init__(self, session: Session):
        self._session = session

    def render(
        self,
        only: Optional[Iterable[str]] = None,
        program_name: str = "pydbc_socketcan_app",
    ) -> str:
        msgs = _collect_messages(self._session, only)
        ns: Dict[str, Any] = {
            "program_name": program_name,
            "messages": msgs,
        }
        text = (
            self.TEMPLATE.decode("utf-8")
            if isinstance(self.TEMPLATE, (bytes, bytearray))
            else str(self.TEMPLATE)
        )
        return renderTemplateFromText(text, ns, encoding="utf-8")


class PythonCanAppGenerator:
    """Generates eine standalone Python-Anwendung (.py) auf Basis von python-can.

    Nutzung:
        gen = PythonCanAppGenerator(session)
        code = gen.render(only=["EngineData"], app_name="pydbc_python_can_app")
        Path("python_can_app.py").write_text(code, encoding="utf-8")
    """

    TEMPLATE = pkgutil.get_data("pydbc", "cgen/templates/python_can.py.tmpl")

    def __init__(self, session: Session):
        self._session = session

    def render(
        self,
        only: Optional[Iterable[str]] = None,
        app_name: str = "pydbc_python_can_app",
    ) -> str:
        msgs = _collect_messages(self._session, only)
        ns: Dict[str, Any] = {
            "app_name": app_name,
            "messages": msgs,
        }
        text = (
            self.TEMPLATE.decode("utf-8")
            if isinstance(self.TEMPLATE, (bytes, bytearray))
            else str(self.TEMPLATE)
        )
        return renderTemplateFromText(text, ns, encoding="utf-8")


class LinGenericGenerator:
    """Generiert generischen LIN-Code (Python + C++) für Encode/Decode, Scheduling-Metadaten.

    Nutzung:
        gen = LinGenericGenerator(session)
        py_code = gen.render_python(module_name="lin_generated")
        cpp_code = gen.render_cpp(header_name="lin_generated.hpp")
    """

    TEMPLATE_PY = pkgutil.get_data("pydbc", "cgen/templates/lin_generic.py.tmpl")
    TEMPLATE_HPP = pkgutil.get_data("pydbc", "cgen/templates/lin_generic.hpp.tmpl")

    def __init__(self, session: Session):
        self._session = session

    def render_python(self, module_name: str = "lin_generated") -> str:
        data = _collect_lin_data(self._session)
        ns: Dict[str, Any] = {
            "module_name": module_name,
            **data,
        }
        text = (
            self.TEMPLATE_PY.decode("utf-8")
            if isinstance(self.TEMPLATE_PY, (bytes, bytearray))
            else str(self.TEMPLATE_PY)
        )
        return renderTemplateFromText(text, ns, encoding="utf-8")

    def render_cpp(self, header_name: str = "lin_generated.hpp") -> str:
        data = _collect_lin_data(self._session)
        ns: Dict[str, Any] = {
            "header_name": header_name,
            **data,
        }
        text = (
            self.TEMPLATE_HPP.decode("utf-8")
            if isinstance(self.TEMPLATE_HPP, (bytes, bytearray))
            else str(self.TEMPLATE_HPP)
        )
        return renderTemplateFromText(text, ns, encoding="utf-8")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2025 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"

import mmap
import os
import re
import sqlite3
from functools import partial

from sqlalchemy import (
    types,
    orm,
    event,
    create_engine,
    text,
)
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

from pydbc.db import model
from pydbc.logger import Logger

DB_EXTENSION = "vndb"

CACHE_SIZE = 4  # MB
PAGE_SIZE = mmap.PAGESIZE


def calculateCacheSize(value):
    return -(value // PAGE_SIZE)


REGEXER_CACHE = {}


def regexer(value, expr):
    if not REGEXER_CACHE.get(expr):
        REGEXER_CACHE[expr] = re.compile(expr, re.UNICODE)
    re_expr = REGEXER_CACHE[expr]
    return re_expr.match(value) is not None


INITIAL_DATA = {
    "node": (
        {
            "rid": 0,
            "node_id": 0,
            "name": "Vector__XXX",
            "comment": "Dummy node for non-existent senders/receivers.",
            "type_": "Node",
        },
    ),
}
"""
    INSERT INTO message(comment,rid,name,message_id,dlc,sender,type) VALUES(
        'This is a message for not used signals, created by Vector CANdb++ DBC OLE DB Provider.',
        1,
        'VECTOR__INDEPENDENT_SIG_MSG',
        3221225472,
        0,
        0,
        'Message');
"""


def _inserter(data, target, conn, **kws):
    for row in data:
        k, v = row.keys(), row.values()
        keys = ", ".join([x for x in k])
        values = ", ".join([repr(x) for x in v])
        stmt = text("INSERT INTO {}({}) VALUES ({})".format(target.name, keys, values))
        conn.execute(stmt)


def loadInitialData(target):
    data = INITIAL_DATA[target.__table__.fullname]
    event.listen(target.__table__, "after_create", partial(_inserter, data))


class MyCustomEnum(types.TypeDecorator):
    impl = types.Integer

    def __init__(self, enum_values, *l, **kw):
        types.TypeDecorator.__init__(self, *l, **kw)
        self._enum_values = enum_values

    def convert_bind_param(self, value, engine):
        result = self.impl.convert_bind_param(value, engine)
        if result not in self._enum_values:
            raise TypeError("Value %s must be one of %s" % (result, self._enum_values))
        return result

    def convert_result_value(self, value, engine):
        "Do nothing here"
        return self.impl.convert_result_value(value, engine)


@event.listens_for(Engine, "connect")
def set_sqlite3_pragmas(dbapi_connection, connection_record):
    dbapi_connection.create_function("REGEXP", 2, regexer)
    cursor = dbapi_connection.cursor()
    # cursor.execute("PRAGMA jornal_mode=WAL")
    cursor.execute("PRAGMA FOREIGN_KEYS=ON")
    cursor.execute("PRAGMA PAGE_SIZE={}".format(PAGE_SIZE))
    cursor.execute(
        "PRAGMA CACHE_SIZE={}".format(calculateCacheSize(CACHE_SIZE * 1024 * 1024))
    )
    cursor.execute("PRAGMA SYNCHRONOUS=OFF")  # FULL
    cursor.execute("PRAGMA LOCKING_MODE=EXCLUSIVE")  # NORMAL
    cursor.execute("PRAGMA TEMP_STORE=MEMORY")  # FILE
    cursor.close()


class VNDB(object):
    """ """

    def __init__(self, filename: str = ":memory:", debug: bool = False, logLevel: str = "INFO", create: bool = True,
                 autocommit: bool = False):
        if filename == ":memory:":
            self.dbname = ""
        else:
            if not filename.lower().endswith(DB_EXTENSION):
                self.dbname = "{}.{}".format(filename, DB_EXTENSION)
            else:
                self.dbname = filename
        self._engine = create_engine(
            "sqlite:///{}".format(self.dbname),
            echo=debug,
            connect_args={
                "detect_types": sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
                "isolation_level": "IMMEDIATE",
            },
            native_datetime=True,
            # autocommit=autocommit
        )
        SessionFactory = sessionmaker(self._engine, autoflush=False)    # autocommit=autocommit
        self._session = SessionFactory()
        self._metadata = model.Base.metadata
        if create == True:
            model.Base.metadata.create_all(self.engine)
            self.session.flush()
            self.session.commit()
            # Mark database as consistent after successful import
            from pydbc.db.model import VndbMeta
            self.session.add(VndbMeta(schema_version=1, vndb_type=1))
            self.session.commit()
        self.logger = Logger(__name__, level=logLevel)

    @classmethod
    def _open_or_create(
            cls, filename=":memory:", debug=False, logLevel="INFO", create=True, autocommit=False,
    ):
        """ """
        inst = cls(filename, debug, logLevel, create, autocommit)
        return inst

    @classmethod
    def create(cls, filename=":memory:", debug=False, logLevel="INFO", autocommit: bool=False):
        """ """
        return cls._open_or_create(filename, debug, logLevel, True, autocommit)

    @classmethod
    def open(cls, filename=":memory:", debug=False, logLevel="INFO", autocommit: bool=False):
        """ """
        return cls._open_or_create(filename, debug, logLevel, False, autocommit)

    @classmethod
    def open_or_create(cls, db_path: str, import_func, debug=False, logLevel="INFO"):
        """
        Opens a VNDB database if it exists and is consistent.
        If not, it deletes the existing file (if any) and recreates it by calling import_func.

        :param db_path: Path to the .vndb database file.
        :param import_func: A callable (e.g., a lambda) that performs the import and returns a SQLAlchemy session.
        :param debug: Enable SQLAlchemy debug logging.
        :param logLevel: Logging level.
        :return: An instance of VNDB.
        """
        if os.path.exists(db_path):
            try:
                db = cls(db_path, debug=debug, logLevel=logLevel, create=False)
                # Check for consistency
                from pydbc.db.model import VndbMeta
                result = db.session.query(VndbMeta).first() # filter_by(key='db_status', value='consistent')
                if result:
                    db.logger.info(f"Opened existing database: {db_path}")
                    return db
                else:
                    db.logger.warning(f"Database '{db_path}' is not consistent. Re-importing.")
                    db.session.close()  # Close connection before deleting
            except Exception as e:
                # This can happen if the file is not a valid SQLite DB
                Logger(__name__).warning(f"Failed to open '{db_path}' as a database ({e}). Re-importing.")

            # If we are here, the DB is inconsistent or corrupt, so we remove it
            try:
                os.remove(db_path)
            except OSError as e:
                Logger(__name__).error(f"Error removing inconsistent database '{db_path}': {e}")
                raise

        # If the file doesn't exist or was removed, import it
        Logger(__name__).info(f"Creating new database by importing source for '{db_path}'...")
        session = import_func()
        return cls.from_session(session)

    @classmethod
    def from_session(cls, session):
        """Creates a VNDB instance from an existing session."""
        instance = cls.__new__(cls)
        instance._session = session
        instance._engine = session.get_bind()
        instance.dbname = str(instance._engine.url).replace("sqlite:///", "")
        instance._metadata = model.Base.metadata
        instance.logger = Logger(__name__)
        return instance

    def close(self):
        """ """
        self.session.close()
        self.engine.dispose()

    @property
    def engine(self):
        return self._engine

    @property
    def metadata(self):
        return self._metadata

    @property
    def session(self):
        return self._session

    def begin_transaction(self):
        """ """

    def commit_transaction(self):
        """ """

    def rollback_transaction(self):
        """ """


loadInitialData(model.Node)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2021 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__  = 'Christoph Schueler'
__version__ = '0.1.0'

from collections import namedtuple

from sqlalchemy.sql.expression import literal, bindparam

from pydbc.logger import Logger
from pydbc import parser
from pydbc.types import AttributeType, BusType, CategoryType, ValueType


from pydbc.db.model import (
    Vndb_Protocol
)

from pydbc.api.ldf import (
    LinMasterNode, LinNetwork, LinNode, LinSlaveNode,
    LinResponseErrorSignal, LinSignal, LinSignalEncodingType, LinSignalEncodingEntry_Value, LinSignalEncodingEntry_Logical,
    LinSignalEncodingEntry_Physical, LinSignalRepresentation, LinScheduleTable, LinScheduleTable_Command_Frame,
    LinScheduleTable_Command_MasterReq, LinScheduleTable_Command_SlaveResp, LinScheduleTable_Command_AssignNad,
    LinScheduleTable_Command_ConditionalChangeNad, LinScheduleTable_Command_DataDump, LinScheduleTable_Command_SaveConfiguration,
    LinScheduleTable_Command_AssignFrameIdRange, LinScheduleTable_Command_FreeFormat,
    LinScheduleTable_Command_AssignFrameId, LinSporadicFrame, LinUnconditionalFrame, LinEventTriggeredFrame,
    LinConfigurableFrame, LinFaultStateSignal,
)

class AttributeContainer:

    def __init__(self, name, objType, valueType, minimum = None, maximum = None, array = False):
        self.name = name
        self.objType = objType
        self.valueType = valueType
        self.minimum = minimum
        self.maximum = maximum
        self.array = array

    def __repr__(self):
        return "{}({} {} {} {} {} {})".format(self.__class__.__name__, self.name, self.objType, self.valueType, self.minimum, self.maximum, self.array)


LDF_ATTRS = {
    # Network
    "LIN_protocol_version": AttributeContainer("LIN_protocol_version", AttributeType.NETWORK, ValueType.STRING),
    "LIN_language_version": AttributeContainer("LIN_language_version", AttributeType.NETWORK, ValueType.STRING),
    "LIN_file_revision": AttributeContainer("LIN_file_revision", AttributeType.NETWORK, ValueType.STRING),
    "LIN_speed": AttributeContainer("LIN_speed", AttributeType.NETWORK, ValueType.FLOAT),
    "LIN_channel_name": AttributeContainer("LIN_channel_name", AttributeType.NETWORK, ValueType.STRING),
    "LIN_include_diag": AttributeContainer("LIN_include_diag", AttributeType.NETWORK, ValueType.INT),
    # Nodes
    "LIN_is_master": AttributeContainer("LIN_is_master", AttributeType.NODE, ValueType.INT),
    "LIN_is_slave": AttributeContainer("LIN_is_slave", AttributeType.NODE, ValueType.INT),
    "LIN_time_base": AttributeContainer("LIN_time_base", AttributeType.NODE, ValueType.FLOAT),
    "LIN_jitter": AttributeContainer("LIN_jitter", AttributeType.NODE, ValueType.FLOAT),
    "LIN_protocol": AttributeContainer("LIN_protocol", AttributeType.NODE, ValueType.STRING),
    "LIN_configured_NAD": AttributeContainer("LIN_configured_NAD", AttributeType.NODE, ValueType.INT),
    "LIN_initial_NAD": AttributeContainer("LIN_initial_NAD", AttributeType.NODE, ValueType.INT),
    "LIN_supplier_id": AttributeContainer("LIN_supplier_id", AttributeType.NODE, ValueType.INT),
    "LIN_function_id": AttributeContainer("LIN_function_id", AttributeType.NODE, ValueType.INT),
    "LIN_variant": AttributeContainer("LIN_variant", AttributeType.NODE, ValueType.INT),
    "LIN_response_error": AttributeContainer("LIN_response_error", AttributeType.NODE, ValueType.STRING),
    #fault_state_signals*        identifierValue [SIGNAL]
    "LIN_P2_min": AttributeContainer("LIN_P2_min", AttributeType.NODE, ValueType.FLOAT),
    "LIN_ST_min": AttributeContainer("LIN_ST_min", AttributeType.NODE, ValueType.FLOAT),
    "LIN_N_As_timeout": AttributeContainer("LIN_N_As_timeout", AttributeType.NODE, ValueType.FLOAT),
    "LIN_N_Cr_timeout": AttributeContainer("LIN_N_Cr_timeout", AttributeType.NODE, ValueType.FLOAT),
    # Signals
    "LIN_signal_initial_value": AttributeContainer("LIN_signal_initial_value", AttributeType.SIGNAL, ValueType.INT, array = True),
}


class LdfListener(parser.BaseListener):

    def __init__(self, database, logLevel = 'INFO', *args, **kws):
        super(LdfListener, self).__init__(database, logLevel, *args, **kws)
        self.logger = Logger(__name__, level = logLevel)
        self.nodes = {}
        self.signals = {}
        self.frames = {}
        self.event_trigged_frames = []
        self.bake_queries()
        self.insertNetwork()

    def bake_queries(self):
        self.UNCONDITIONAL_FRAME_BY_NAME = self.bakery(lambda session: session.query(LinUnconditionalFrame).filter(
            LinUnconditionalFrame.name == bindparam("name")))
        self.SCHEDULE_TABLE_BY_NAME = self.bakery(lambda session: session.query(LinScheduleTable).filter(
            LinScheduleTable.name == bindparam('name')))
        self.ENCODING_TYPE_BY_NAME = self.bakery(lambda session: session.query(LinSignalEncodingType).filter(
            LinSignalEncodingType.name == bindparam("name")))
        self.LIN_NODE_BY_NAME = self.bakery(lambda session: session.query(LinNode).filter(LinNode.name == bindparam('name')))
        self.LIN_SIGNAL_BY_NAME = self.bakery(lambda session: session.query(LinSignal).filter(LinSignal.name == bindparam('name')))
        self.LIN_SIGNAL_BY_RID = self.bakery(lambda session: session.query(LinSignal).filter(LinSignal.rid == bindparam('rid')))

    def insertNetwork(self, specific = None):
        self.log_insertion("Network")
        network = LinNetwork(name = self.db.dbname)
        self.session.add(network)
        proto = Vndb_Protocol(network = network, name = BusType.LIN.name, specific = specific)
        self.session.add(proto)
        self.session.flush()
        self.network_id = network.lin_network_id

    def updateNetworkAttributes(self):
        value = self.value
        languageVersion = value['languageVersion']
        protocolVersion = value['protocolVersion']
        fileRevision = value['fileRevision']
        speed = value['speed']
        channelName = value['channelName']
        network = self.session.query(LinNetwork).filter(LinNetwork.lin_network_id == self.network_id).first()
        network.language_version = languageVersion
        network.protocol_version = protocolVersion
        network.file_revision = fileRevision
        network.speed = speed
        network.channel_name = channelName
        self.session.flush()

    def insertConfigurableFrames(self):
        self.log_insertion("ConfigurableFrames")
        for node, frames in self.configurableFrames.items():
            for frameName, messageID in frames:
                frame = self.UNCONDITIONAL_FRAME_BY_NAME(self.session).params(name = frameName).first()
                if not frame:
                    self.logger.error("While inserting configurable frames: frame '{}' does not exist.".format(frameName))
                    continue
                cf = LinConfigurableFrame(node = node, frame = frame, identifier = messageID)
                self.session.add(cf)
        self.session.flush()

    def insertFaultStateSignals(self):
        self.log_insertion("FaultStateSignals")
        for node, signals in self.faultStateSignals.items():
            for signalName in signals:
                signal = self.LIN_SIGNAL_BY_NAME(self.session).params(name = signalName).first()
                if not signal:
                    self.logger.error("While inserting fault state signals: signal '{}' does not exist.".format(signalName))
                    continue
                lfs = LinFaultStateSignal(node = node, signal = signal)
                self.session.add(lfs)
        self.session.flush()

    def insertResponseErrorSignals(self):
        self.log_insertion("ResponseErrorSignals")
        for node, signalName in self.responseErrorSignals.items():
            signal = self.LIN_SIGNAL_BY_NAME(self.session).params(name = signalName).first()
            if not signal:
                self.logger.error("While inserting response error signals: signal '{}' does not exist.".format(signalName))
                continue
            res = LinResponseErrorSignal(node = node, signal = signal)
            self.session.add(res)
        self.session.flush()

    def sendingFrameNames(self, signal):
        result = []
        for _, frame in self.frames.items():
            frameSignals = (s['signalName'] for s in frame['signals'])
            frameName = frame['frameName']
            if signal in frameSignals:
                result.append(frameName)
        return result

    def updateEventTriggeredFrames(self):
        for etf, name in self.event_trigged_frames:
            lst = self.SCHEDULE_TABLE_BY_NAME(self.session).params(name = name).first()
            if not lst:
                self.logger.error("While updating event triggered frames: schedult table '{}' does not exist.".format(name))
                continue
            etf.collision_resolving_schedule_table = lst
        self.session.flush()

    def exitLin_description_file(self, ctx):
        self.value = dict(
            protocolVersion = ctx.pv.value,
            languageVersion = ctx.lv.value,
            fileRevision = ctx.fr.value if ctx.fr else None,
            speed = ctx.ls.value,
            channelName = ctx.cn.value if ctx.cn else None,
            nodes = ctx.ndef.value,
            nodeCompositions = ctx.ncdef.value if ctx.ncdef else None,
            signals = ctx.sdef.value,
            diagnosticSignals = ctx.dsdef.value if ctx.dsdef else None,
            frames = ctx.fdef.value,
            sporadicFrames = ctx.sfdef.value if ctx.sfdef else [],
            eventTriggeredFrames = ctx.etfdef.value if ctx.etfdef else [],
            diagnosticFrames = ctx.dffdef.value if ctx.dffdef else [],
            nodeAttributes = ctx.nadef.value,
            scheduleTables = ctx.stdef.value,
            signalGroups = ctx.sgdef.value if ctx.sgdef else [],
            signalEncodings = ctx.setdef.value if ctx.setdef else [],
            signalRepresentations = ctx.srdef.value if ctx.srdef else [],
        )
        self.updateNetworkAttributes()
        self.insertConfigurableFrames()
        self.insertFaultStateSignals()
        self.insertResponseErrorSignals()
        self.updateEventTriggeredFrames()
        self.session.commit()

    def exitLin_protocol_version_def(self, ctx):
        ctx.value = self.getValue(ctx.s)

    def exitLin_language_version_def(self, ctx):
        ctx.value = self.getValue(ctx.s)

    def exitLin_file_revision_def(self, ctx):
        ctx.value = self.getValue(ctx.s)

    def exitLin_speed_def(self, ctx):
        ctx.value = self.getValue(ctx.n)

    def exitChannel_name_def(self, ctx):
        ctx.value = self.getValue(ctx.i)

    def exitNode_def(self, ctx):
        self.log_insertion("Nodes")
        mname = self.getValue(ctx.mname)
        timebase = self.getValue(ctx.tb)
        jitter = self.getValue(ctx.j)
        bit_length = self.getValue(ctx.bit_length)
        tolerant = self.getValue(ctx.tolerant)
        snames = [x.value for x in ctx.snames]
        ctx.value = dict(
            master = mname, timeBase = timebase, jitter = jitter, slaves = snames, bit_length = bit_length, tolerant = tolerant
        )
        nodes = ctx.value
        masterNode = nodes['master']
        master = LinMasterNode(name = masterNode, timebase = timebase, jitter = jitter, bit_length = bit_length, tolerant = tolerant)
        self.session.add(master)
        self.master_node = master
        self.session.flush()
        mrid = master.rid
        self.nodes[masterNode] = master
        self.master = masterNode
        for name in nodes['slaves']:
            slave = LinSlaveNode(name = name)
            self.session.add(slave)
            self.session.flush()    # ???
            srid = slave.lin_node_id
            self.nodes[name] = slave
        self.session.flush()

    def exitNode_attributes_def(self, ctx):
        self.log_insertion("NodeAttributes")
        items = [x.value for x in ctx.items]
        ctx.value = items
        attrs = ctx.value
        self.configurableFrames = {}
        self.faultStateSignals = {}
        self.responseErrorSignals = {}
        for attr in attrs:
            #print("ATTR:", attr)
            name = attr['name']
            node = self.nodes.get(name)
            if not node:
                self.logger.error("While inserting attribute definition: node '{}' does not exist.".format(name))
                continue
            node.protocol_version = attr['version']
            node.configured_NAD = attr['configuredNAD']
            node.initial_NAD = attr['initialNAD']
            node.supplier_id = attr['supplierID']
            node.function_id = attr['functionID']
            node.variant = attr['variant']
            node.p2_min = attr['p2Min']
            node.st_min = attr['stMin']
            node.n_as_timeout = attr['nAs']
            node.n_cr_timeout = attr['nCr']
            node.response_tolerance = attr['response_tolerance']
            if 'configurableFrames' in attr:
                self.configurableFrames[node] = []
                for cframe in attr['configurableFrames']:
                    self.configurableFrames[node].append((cframe['frameName'],cframe['messageID'], ))
            if 'faultStateSignals' in attr:
                self.faultStateSignals[node] = []
                for fs in attr['faultStateSignals']:
                    self.faultStateSignals[node].append(fs)
            if 'responseErrorSignal' in attr:
                res = attr['responseErrorSignal']
                self.responseErrorSignals[node] = res

        self.session.flush()

    def exitNode_attribute(self, ctx):
        name = self.getValue(ctx.name)
        version = self.getValue(ctx.version)
        n0 = self.getValue(ctx.n0)
        n1 = self.getValue(ctx.n1)
        attrs = self.getValue(ctx.attrs, dict())
        ctx.value = dict(name = name, version = version, configuredNAD = n0, initialNAD = n1, **attrs)

    def exitAttributes_def(self, ctx):
        sid = self.getValue(ctx.sid)
        fid = self.getValue(ctx.fid)
        v = self.getValue(ctx.v)
        sn0 = self.getValue(ctx.sn0)
        sn1s = [x.value for x in ctx.sn1s]
        cf = self.getValue(ctx.cf)
        p2Min = self.getValue(ctx.p2Min)
        stMin = self.getValue(ctx.stMin)
        nAs = self.getValue(ctx.nAs)
        nCr = self.getValue(ctx.nCr)
        response_tolerance = self.getValue(ctx.response_tolerance)
        ctx.value = dict(supplierID = sid, functionID = fid, variant = v, responseErrorSignal = sn0, faultStateSignals = sn1s,
            p2Min = p2Min, stMin = stMin, nAs = nAs, nCr = nCr, configurableFrames = cf, response_tolerance = response_tolerance
        )

    def exitConfigurable_frames(self, ctx):
        ctx.value = [x.value for x in ctx.frames]

    def exitConfigurable_frame(self, ctx):
        fname = self.getValue(ctx.fname)
        mid = self.getValue(ctx.mid)
        ctx.value = dict(frameName = fname, messageID = mid)

    def exitNode_composition_def(self, ctx):
        ctx.value = [x.value for x in ctx.items]

    def exitConfiguration(self, ctx):
        cname = self.getValue(ctx.cname)
        items = [x.value for x in ctx.items]
        ctx.value = dict(configurationName = cname, items = items)

    def exitConfiguration_item(self, ctx):
        cnode = self.getValue(ctx.cnode)
        lnodes = [x.value for x in ctx.lnodes]
        ctx.value = dict(compositeNode = cnode, logicalNodes = lnodes)

    def exitSignal_def(self, ctx):
        self.log_insertion("Signals")
        ctx.value = [x.value for x in ctx.items]
        for signal in ctx.value:
            #print("SIG:", signal)
            initValue = signal['initValue']
            name = signal['name']
            size = signal['size']
            publishedBy = signal['publishedBy']
            subscribedBy = signal['subscribedBy']
            publisher = self.nodes.get(publishedBy)
            if not publisher:
                self.logger.error("While inserting signals: publisher node '{}' does not exist.".format(publishedBy))
                continue
            if not initValue['scalar'] is None:
                iv = initValue['scalar']
            elif not initValue['array'] is None:
                iv = initValue['array']
            else:
                self.logger.error("While inserting signals: no initial value for signal '{}'.".format(name))
                iv = None
            sig = LinSignal(name = name, signal_size = size, init_value = iv, publisher = publisher)
            self.session.add(sig)

            for subscriber in subscribedBy:
                node = self.nodes.get(subscriber)
                if not node:
                    self.logger.error("While inserting signals: subscriber node '{}' does not exist.".format(subscriber))
                    continue
                sig.subscribers.append(node)
            self.session.flush()
            signal['rid'] = sig.rid
            self.signals[name] = signal
        #self.session.flush()

    def exitSignal_item(self, ctx):
        sname = self.getValue(ctx.sname)
        ssize = self.getValue(ctx.ssize)
        initValue = self.getValue(ctx.initValue)
        pub = self.getValue(ctx.pub)
        sub = [x.value for x in ctx.sub]
        ctx.value = dict(name = sname, size = ssize, initValue = initValue, publishedBy = pub, subscribedBy = sub)

    def exitInit_value(self, ctx):
        scalar = self.getValue(ctx.s)
        array = self.getValue(ctx.a)
        ctx.value = dict(scalar = scalar, array = array)

    def exitInit_value_scalar(self, ctx):
        ctx.value = self.getValue(ctx.i)

    def exitInit_value_array(self, ctx):
        ctx.value = [x.value for x in ctx.vs]

    def exitDiagnostic_signal_def(self, ctx):
        ctx.value = [x.value for x in ctx.items]

    def exitDiagnostic_item(self, ctx):
        name = self.getValue(ctx.name)
        size = self.getValue(ctx.size)
        initValue = self.getValue(ctx.initValue)
        ctx.value = dict(name = name, size = size, initValue = initValue)

    def exitSignal_groups_def(self, ctx):
        ctx.value = [x.value for x in ctx.items]

    def exitSignal_group(self, ctx):
        sgname = self.getValue(ctx.sgname)
        gsize = self.getValue(ctx.gsize)
        items = [x.value for x in ctx.items]
        ctx.value = dict(signalGroupName = sgname, groupSize = gsize, items = items)

    def exitSignal_group_item(self, ctx):
        sname = self.getValue(ctx.sname)
        goffs = self.getValue(ctx.goffs)
        ctx.value = dict(signalName = sname, groupOffset = goffs)

    def exitFrame_def(self, ctx):
        self.log_insertion("UnconditionalFrames")
        ctx.value = [x.value for x in ctx.items]
        for frame in ctx.value:
            frid = frame['frameID']
            name = frame['frameName']
            size = frame['frameSize']
            publishedBy = frame['publishedBy']
            publisher = self.nodes.get(publishedBy)
            if not publisher:
                self.logger.error("While inserting unconditional frames: node '{}' does not exist.".format(publishedBy))
                continue
            msg = LinUnconditionalFrame(name = name, frame_id = frid, size = size, publisher = publisher)
            self.session.add(msg)
            self.session.flush()
            #print("FRAME:", msg)
            frame['rid'] = msg.rid
            self.frames[name] = frame
            for signal in frame['signals']:
                signalOffs = signal['signalOffset']
                signalName = signal['signalName']
                srid = self.signals[signalName]['rid']
                sgn = self.LIN_SIGNAL_BY_RID(self.session).params(rid = srid).first()
                if not sgn:
                    self.logger.error("While inserting unconditional frames: signal '{}' does not exist.".format(srid))
                    continue
                msg.add_signal(sgn, signalOffs)
        self.session.flush()

    def exitFrame_item(self, ctx):
        fname = self.getValue(ctx.fname)
        fid = self.getValue(ctx.fid)
        p = self.getValue(ctx.p)
        fsize = self.getValue(ctx.fsize)
        items = [x.value for x in ctx.items]
        ctx.value = dict(frameName = fname, frameID = fid, publishedBy = p, frameSize = fsize, signals = items)

    def exitFrame_signal(self, ctx):
        sname = self.getValue(ctx.sname)
        soffs = self.getValue(ctx.soffs)
        ctx.value = dict(signalName = sname, signalOffset = soffs)

    def exitSporadic_frame_def(self, ctx):
        self.log_insertion("SporadicFrames")
        ctx.value = [x.value for x in ctx.items]
        for sf in ctx.value:
            #print("*SPOR", sf)
            name = sf['sporadicFrameName']
            lsf = LinSporadicFrame(name = name)
            self.session.add(lsf)
            self.session.flush()
            for frameName in sf['frameNames']:
                frame = self.MESSAGE_BY_NAME(self.session).params(name = frameName).first()
                if not frame:
                    self.logger.error("While inserting sporadic frames: frame '{}' does not exist.".format(frameName))
                    continue
                lsf.associated_frames.append(frame)
        self.session.flush()

    def exitSporadic_frame_item(self, ctx):
        name = ctx.sfn.value
        fnames = [x.value for x in ctx.names]
        ctx.value = dict(sporadicFrameName = name, frameNames = fnames)

    def exitEvent_triggered_frame_def(self, ctx):
        self.log_insertion("EventTriggeredFrames")
        ctx.value = [x.value for x in ctx.items]
        for ef in ctx.value:
            name = ef['frameName']
            scheduleTable = ef['scheduleTable']
            frameID = ef['frameID']
            etf = LinEventTriggeredFrame(name = name, master_node = self.master_node, frame_id = frameID)
            self.session.add(etf)
            self.event_trigged_frames.append((etf, scheduleTable))
            for frameName in ef['frameNames']:
                frame = self.MESSAGE_BY_NAME(self.session).params(name = frameName).first()
                if not frame:
                    self.logger.error("While inserting event triggered frames: frame '{}' does not exist.".format(frameName))
                    continue
                etf.associated_frames.append(frame)
        self.session.flush()

    def exitEvent_triggered_frame_item(self, ctx):
        e = self.getValue(ctx.e)
        c = self.getValue(ctx.c)
        fid = self.getValue(ctx.fid)
        items = [x.value for x in ctx.items]
        ctx.value = dict(frameName = e, frameID = fid, scheduleTable = c, frameNames = items)

    def exitDiag_frame_def(self, ctx):
        mid = self.getValue(ctx.mid)
        sid = self.getValue(ctx.sid)
        mitems = [x.value for x in ctx.mitems]
        sitems = [x.value for x in ctx.sitems]
        ctx.value = dict(masterID = mid, slaveID = sid, masterSignals = mitems, slaveSignals = sitems)

    def exitDiag_frame_item(self, ctx):
        sname = self.getValue(ctx.sname)
        soffs = self.getValue(ctx.soffs)
        ctx.value = dict(signalName = sname, signalOffset = soffs)

    def exitSchedule_table_def(self, ctx):
        self.log_insertion("ScheduleTables")
        ctx.value = [x.value for x in ctx.items]
        for table in ctx.value:
            name = table['name']
            lst = LinScheduleTable(name = name)
            self.session.add(lst)
            self.session.flush()
            for cmd in table['commands']:
                frame_time = cmd['frameTime']
                cmd = cmd['command']
                ct = cmd['type']
                if ct == 'Frame':
                    name = cmd['frame_name']
                    frame = self.MESSAGE_BY_NAME(self.session).params(name = name).first()
                    if not frame:
                        self.logger.error("While inserting schedule tables: frame '{}' does not exist.".format(name))
                        continue
                    entry = LinScheduleTable_Command_Frame(frame_time = frame_time, frame = frame)
                elif ct == 'MasterReq':
                    entry = LinScheduleTable_Command_MasterReq(frame_time = frame_time)
                elif ct == 'SlaveResp':
                    entry = LinScheduleTable_Command_SlaveResp(frame_time = frame_time)
                elif ct == 'AssignNAD':
                    node_name = cmd['node_name']
                    node = self.nodes.get(node_name)
                    if not node:
                        self.logger.error("While inserting schedule tables: node '{}' does not exist.".format(name))
                        continue
                    entry = LinScheduleTable_Command_AssignNad(frame_time = frame_time, node = node)
                elif ct == 'ConditionalChangeNAD':
                    nad = cmd['nad']
                    id_ = cmd['id']
                    byte_ = cmd['byte']
                    mask = cmd['mask']
                    inv = cmd['inv']
                    new_nad = cmd['new_nad']
                    entry = LinScheduleTable_Command_ConditionalChangeNad(frame_time = frame_time, nad = nad,
                        id = id_, byte = byte_, mask = mask, inv = inv, new_nad = new_nad
                    )
                elif ct == 'DataDump':
                    node_name = cmd['node_name']
                    node = self.nodes.get(node_name)
                    if not node:
                        self.logger.error("While inserting schedule tables: node '{}' does not exist.".format(name))
                        continue
                    d1 = cmd['d1']
                    d2 = cmd['d2']
                    d3 = cmd['d3']
                    d4 = cmd['d4']
                    d5 = cmd['d5']
                    entry = LinScheduleTable_Command_DataDump(frame_time = frame_time, node = node, d1 = d1,
                        d2 = d2, d3 = d3, d4 = d4, d5 = d5
                    )
                elif ct == 'SaveConfiguration':
                    node_name = cmd['node_name']
                    node = self.nodes.get(node_name)
                    if not node:
                        self.logger.error("While inserting schedule tables: node '{}' does not exist.".format(name))
                        continue
                    entry = LinScheduleTable_Command_SaveConfiguration(frame_time = frame_time, node = node)
                elif ct == 'AssignFrameIdRange':
                    node_name = cmd['node_name']
                    node = self.nodes.get(node_name)
                    if not node:
                        self.logger.error("While inserting schedule tables: node '{}' does not exist.".format(name))
                        continue
                    frame_index = cmd['frame_index']
                    frame_pid1 = cmd['pid1']
                    frame_pid2 = cmd['pid2']
                    frame_pid3 = cmd['pid3']
                    frame_pid4 = cmd['pid4']
                    entry = LinScheduleTable_Command_AssignFrameIdRange(frame_time = frame_time, node = node,
                        frame_index = frame_index, frame_pid1 = frame_pid1, frame_pid2 = frame_pid2,
                        frame_pid3 = frame_pid3, frame_pid4 = frame_pid4
                    )
                elif ct == 'FreeFormat':
                    d1 = cmd['d1']
                    d2 = cmd['d2']
                    d3 = cmd['d3']
                    d4 = cmd['d4']
                    d5 = cmd['d5']
                    d6 = cmd['d6']
                    d7 = cmd['d7']
                    d8 = cmd['d8']
                    entry = LinScheduleTable_Command_FreeFormat(frame_time = frame_time, d1 = d1, d2 = d2, d3 = d3,
                        d4 = d4, d5 = d5, d6 = d6, d7 = d7, d8 = d8
                    )
                elif ct == 'AssignFrameId':
                    node_name = cmd['nodeName']
                    frame_name = cmd['frameName']
                    node = self.nodes.get(node_name)
                    if not node:
                        self.logger.error("While inserting schedule tables: node '{}' does not exist.".format(name))
                        continue
                    frame = self.MESSAGE_BY_NAME(self.session).params(name = frame_name).first()
                    if not frame:
                        self.logger.error("While inserting schedule tables: frame '{}' does not exist.".format(frame_name))
                        continue
                    entry = LinScheduleTable_Command_AssignFrameId(frame_time = frame_time, node = node, frame = frame)
                self.session.add(entry)
                lst.entries.append(entry)
            self.session.flush()

    def exitSchedule_table_entry(self, ctx):
        s = self.getValue(ctx.s)
        items = [x.value for x in ctx.items]
        ctx.value = dict(name = s, commands = items)

    def exitSchedule_table_command(self, ctx):
        c = self.getValue(ctx.c)
        f = self.getValue(ctx.f)
        ctx.value = dict(command = c, frameTime = f)

    def exitCommand(self, ctx):
        if ctx.frameName:
            cmdName = "Frame"
        else:
            cmdName = ctx.c.text
        cmd = dict(type = cmdName)
        if cmdName == "Frame":
            cmd['frame_name'] = ctx.frameName.value
        elif cmdName == 'AssignNAD':
            cmd['node_name'] = ctx.nodeName.value
        elif cmdName == 'ConditionalChangeNAD':
            cmd['nad'] = ctx.nad.value
            cmd['id'] = ctx.id_.value
            cmd['byte'] = ctx.byte_.value
            cmd['mask'] = ctx.mask.value
            cmd['inv'] = ctx.inv.value
            cmd['new_nad'] = ctx.new_NAD.value
        elif cmdName == 'DataDump':
            cmd['node_name'] = ctx.nodeName.value
            cmd['d1'] = ctx.d1.value
            cmd['d2'] = ctx.d2.value
            cmd['d3'] = ctx.d3.value
            cmd['d4'] = ctx.d4.value
            cmd['d5'] = ctx.d5.value
        elif cmdName == 'SaveConfiguration':
            cmd['node_name'] = ctx.nodeName.value
        elif cmdName == 'AssignFrameIdRange':
            cmd['node_name'] = ctx.nodeName.value
            cmd['frame_index'] = ctx.frameIndex.value
            cmd['pid1'] = None
            cmd['pid2'] = None
            cmd['pid3'] = None
            cmd['pid4'] = None
            pids = [p.value for p in ctx.pids]
            lp = len(pids)
            if lp == 1:
                cmd['pid1'] = pids[0]
            elif lp == 2:
                cmd['pid1'] = pids[0]
                cmd['pid2'] = pids[1]
            elif lp == 3:
                cmd['pid1'] = pids[0]
                cmd['pid2'] = pids[1]
                cmd['pid3'] = pids[2]
            elif lp == 4:
                cmd['pid1'] = pids[0]
                cmd['pid2'] = pids[1]
                cmd['pid3'] = pids[2]
                cmd['pid4'] = pids[3]
        elif cmdName == 'FreeFormat':
            cmd['d1'] = ctx.d1.value
            cmd['d2'] = ctx.d2.value
            cmd['d3'] = ctx.d3.value
            cmd['d4'] = ctx.d4.value
            cmd['d5'] = ctx.d5.value
            cmd['d6'] = ctx.d6.value
            cmd['d7'] = ctx.d7.value
            cmd['d8'] = ctx.d8.value
        elif cmdName == 'AssignFrameId':
            cmd['nodeName'] = ctx.nodeName.value
            cmd['frameName'] = ctx.frName.value
        ctx.value = cmd

    def exitSignal_encoding_type_def(self, ctx):
        self.log_insertion("SignalEncodings")
        items = [x.value for x in ctx.items]
        ctx.value = items
        tps = set()
        for enc in items:
            lse = LinSignalEncodingType(name = enc['name'])
            self.session.add(lse)
            for value in enc['values']:
                vtype = value['valueType']
                value = value["value"]
                if vtype == "logical":
                    entry = LinSignalEncodingEntry_Logical(signal_value = value['signalValue'], text_info = value['text'])
                elif vtype == "range":
                    min_value = value['min']
                    max_value = value['max']
                    scale = value['scale']
                    offset = value['offset']
                    text = value['text']
                    entry = LinSignalEncodingEntry_Physical(min_value = min_value, max_value = max_value,
                        scale = scale, offset = offset, text_info = text
                    )
                elif vtype == "bcd":
                    entry = LinSignalEncodingEntry_Value(entry_type = 2)
                elif vtype == "ascii":
                    entry = LinSignalEncodingEntry_Value(entry_type = 1)
                lse.entries.append(entry)
                self.session.add(entry)
        self.session.flush()

    def exitSignal_encoding_entry(self, ctx):
        s = self.getValue(ctx.s)
        items = [x.value for x in ctx.items]
        ctx.value = dict(name = s, values = items)


    def exitSignal_encoding_value(self, ctx):
        if ctx.l:
            value = self.getValue(ctx.l)
            vtype = "logical"
        elif ctx.p:
            value = self.getValue(ctx.p)
            vtype = "range"
        elif ctx.b:
            value = None
            vtype = "bcd"
        elif ctx.a:
            value = None
            vtype = "ascii"
        ctx.value = dict(value = value, valueType = vtype)

    def exitLogical_value(self, ctx):
        s = self.getValue(ctx.s)
        t = self.getValue(ctx.t)
        ctx.value = dict(signalValue = s, text = t)

    def exitPhysical_range(self, ctx):
        minValue = self.getValue(ctx.minValue)
        maxValue = self.getValue(ctx.maxValue)
        #scale = ctx.scale.value if ctx.scale else None
        #offset = ctx.offset.value
        scale = self.getValue(ctx.s)
        offset = self.getValue(ctx.o)
        t = self.getValue(ctx.t)
        ctx.value = dict(min = minValue, max = maxValue, scale = scale, offset = offset, text = t)

    def exitBcd_value(self, ctx):
        pass

    def exitAscii_value(self, ctx):
        pass

    def exitSignal_representation_def(self, ctx):
        self.log_insertion("SignalRepresentations")
        items = [x.value for x in ctx.items]
        ctx.value = items
        for sr in items:
            name = sr['name']
            lse = self.ENCODING_TYPE_BY_NAME(self.session).params(name = name).first()
            if not lse:
                self.logger.error("While inserting signal representations: encoding type '{}' does not exist.".format(name))
                continue
            for signal_name in sr['signalNames']:
                signal = self.LIN_SIGNAL_BY_NAME(self.session).params(name = signal_name).first()
                if not signal:
                    self.logger.error("While inserting signal representations: signal '{}' does not exist.".format(signal_name))
                    continue
                lsr = LinSignalRepresentation(signal_encoding_type = lse, signal = signal)
                self.session.add(lsr)
        self.session.flush()

    def exitSignal_representation_entry(self, ctx):
        enc = self.getValue(ctx.enc)
        names = [x.value for x in ctx.names]
        ctx.value = dict(name = enc, signalNames = names)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__  = 'Christoph Schueler'
__version__ = '0.1.0'


from sqlalchemy.sql.expression import literal, bindparam

from pydbc.logger import Logger
from pydbc import parser
from pydbc.types import AttributeType, BusType, CategoryType, ValueType
from pydbc.db.model import (
    Dbc_Version, Message, Message_Signal, Network, Node, Signal, Value_Description,
    Valuetable, EnvironmentVariablesData, EnvVar, Attribute_Definition, Attribute_Value,
    Node_TxMessage, Node_RxSignal, Category_Definition, Category_Value, AttributeRel_Value,
    Signal_Group_Signal, Signal_Group, Node_TxSig, LinSignalEncodingType, LinSignalEncodingEntry_Value,
    LinSignalEncodingEntry_Logical, LinSignalEncodingEntry_Physical, LinSignalRepresentation,
    LinScheduleTable, LinScheduleTable_Command_Frame, LinScheduleTable_Command_MasterReq,
    LinScheduleTable_Command_SlaveResp, LinScheduleTable_Command_AssignNad,
    LinScheduleTable_Command_ConditionalChangeNad, LinScheduleTable_Command_DataDump,
    LinScheduleTable_Command_SaveConfiguration, LinScheduleTable_Command_AssignFrameIdRange,
    LinScheduleTable_Command_FreeFormat, LinScheduleTable_Command_AssignFrameId, LinSporadicFrame,
    LinUnconditionalFrame, LinEventTriggeredFrame, LinConfigurableFrame, LinFaultStateSignal,
    Vndb_Protocol
)


class NcfListener(parser.BaseListener):

    def __init__(self, database, logLevel = 'INFO', *args, **kws):
        super(NcfListener, self).__init__(database, logLevel, *args, **kws)
        self.logger = Logger(__name__, level = logLevel)
        #self.bake_queries()

    def exitToplevel(self, ctx):
        v = ctx.v.value
        nodes = [x.value for x in ctx.nodes]
        self.value = dict(version = v, nodes = nodes)

    def exitLanguage_version(self, ctx):
        s = ctx.s.value if ctx.s else None
        ctx.value = s

    def exitNode_definition(self, ctx):
        name = ctx.name.value
        g = ctx.g.value if ctx.g else None
        d = ctx.d.value if ctx.d else None
        f = ctx.f.value if ctx.d else None
        e = ctx.e.value if ctx.e else None
        s = ctx.s.value if ctx.s else None
        t = ctx.t.value if ctx.t else None
        ctx.value = dict(name = name, general = g, diagnostic = d, frames= f, encodings = e, status = s, freeText = t)

    def exitNode_name(self, ctx):
        ctx.value = ctx.i.value

    def exitGeneral_definition(self, ctx):
        pv = ctx.pv.value
        sup = ctx.sup.value
        fun = ctx.fun.value
        var = ctx.var.value
        br = ctx.br.value
        tf = True if ctx.tf.text == "yes" else False
        vfrom = ctx.vfrom.value if ctx.vfrom else None
        vto = ctx.vto.value if ctx.vto else None
        tfrom = ctx.tfrom.value if ctx.tfrom else None
        tto = ctx.tto.value if ctx.tto else None
        conf = ctx.conf.value if ctx.conf else None
        ctx.value = dict(
            protocolVersion = pv, supplier = sup, function = fun, variant = var, bitrate = br, wakeupSignal = tf,
            voltageFrom = vfrom, voltageTo = vto, temperatureFrom = tfrom, temperatureTo = tto, conformance = conf,
        )

    def exitProtocol_version(self, ctx):
        ctx.value = ctx.s.value

    def exitSupplier_id(self, ctx):
        ctx.value = ctx.i.value

    def exitFunction_id(self, ctx):
        ctx.value = ctx.i.value

    def exitVariant_id(self, ctx):
        ctx.value = ctx.i.value

    def exitBitrate_definition(self, ctx):
        rates = br = minBr = maxBr = None
        if ctx.rates:
            rates = [x.value for x in ctx.rates]
        elif ctx.br:
            br = ctx.br.value
        else:
            minBr = ctx.minBr.value
            maxBr = ctx.maxBr.value
        ctx.value = dict(bitrate = br, minBr = minBr, maxBr = maxBr, rates = rates)

    def exitBitrate(self, ctx):
        ctx.value = ctx.n.value

    def exitDiagnostic_definition(self, ctx):
        lhs = ctx.lhs.value
        rhs = ctx.rhs.value if ctx.rhs else None
        nads = [x.value for x in ctx.nads] if ctx.nads else []
        dc = ctx.dc.value if ctx.dc else None
        p2Min = ctx.p2Min.value if ctx.p2Min else None
        stMin = ctx.stMin.value if ctx.p2Min else None
        nAs = ctx.nAs.value if ctx.nAs else None
        nCr = ctx.nCr.value if ctx.nCr else None
        sids = [x.value for x in ctx.sids] if ctx.sids else []
        mml = ctx.mml.value if ctx.mml else None
        ctx.value = dict(
            maxMessageLength = mml, lhs = lhs, rhs = rhs, nads = nads, diagnosticClass = dc, p2Min = p2Min, stMin = stMin, nAs = nAs, nCr = nCr, supportedSids = sids
        )

    def exitFrame_definition(self, ctx):
        frames = [x.value for x in ctx.frames]
        ctx.value = frames

    def exitSingle_frame(self, ctx):
        n = ctx.n.value
        p = ctx.p.value if ctx.p else None
        s = ctx.s.value if ctx.s else None
        ctx.value = dict(name = n, properties = p, signal = s)

    def exitFrame_kind(self, ctx):
        text = ctx.v.text
        ctx.value = text

    def exitFrame_name(self, ctx):
        name = ctx.i.value
        ctx.value = name

    def exitFrame_properties(self, ctx):
        l = ctx.l.value
        minValue = ctx.minValue.value if ctx.minValue else None
        maxValue = ctx.maxValue.value if ctx.maxValue else None
        etf = ctx.etf.value if ctx.etf else None
        ctx.value = dict(length = l, minPeriod = minValue, maxPeriod = maxValue, eventTriggeredFrame = etf)

    def exitSignal_definition(self, ctx):
        ctx.value = [x.value for x in ctx.items]

    def exitSignal_definition_entry(self, ctx):
        n = ctx.n.value
        p = ctx.p.value if ctx.p else None
        ctx.value = dict(name = n, properties = p)

    def exitSignal_name(self, ctx):
        name = ctx.i.value
        ctx.value = name

    def exitSignal_properties(self, ctx):
        init = ctx.init.value
        s = ctx.s.value
        o = ctx.o.value if ctx.o else None
        e = ctx.e.value if ctx.e else None
        ctx.value = dict(initValue = init, size = s, offset = o, encoding = e)

    def exitInit_value(self, ctx):
        scalar = ctx.s.value if ctx.s else None
        array = ctx.a.value if ctx.a else None
        ctx.value = dict(scalar = scalar, array = array)

    def exitInit_value_scalar(self, ctx):
        ctx.value = ctx.i.value

    def exitInit_value_array(self, ctx):
        ctx.value = [x.value for x in ctx.values]

    def exitEncoding_definition(self, ctx):
        ctx.value = [x.value for x in ctx.items]

    def exitEncoding_definition_entry(self, ctx):
        name = ctx.name.value
        items = [x.value for x in ctx.items]
        ctx.value = dict(name = name, values = items)

    def exitEncoding_definition_value(self, ctx):
        if ctx.l:
            value = ctx.l.value
            vtype = "logical"
        elif ctx.p:
            value = ctx.p.value
            vtype = "range"
        elif ctx.b:
            value = None
            vtype = "bcd"
        elif ctx.a:
            value = None
            vtype = "ascii"
        ctx.value = dict(value = value, valueType = vtype)

    def exitSignal_encoding_type_name(self, ctx):
        ctx.value = ctx.i.value

    def exitEncoding_name(self, ctx):
        ctx.value = ctx.i.value

    def exitLogical_value(self, ctx):
        s = ctx.s.value
        t = ctx.t.value if ctx.t else None
        ctx.value = dict(signalValue = s, text = t)

    def exitPhysical_range(self, ctx):
        #'physical_value' ',' minValue = min_value ',' maxValue = max_value ',' s = scale ',' o = offset (',' t = text_info)? ';'
        minValue = ctx.minValue.value
        maxValue = ctx.maxValue.value
        s = ctx.s.value
        o = ctx.o.value
        t = ctx.t.value if ctx.t else None
        ctx.value = dict(min = minValue, max = maxValue, scale = s, offset = o, text = t)

    def exitBcd_value(self, ctx):
        pass

    def exitAscii_value(self, ctx):
        pass

    def exitSignal_value(self, ctx):
        ctx.value = ctx.n.value

    def exitMin_value(self, ctx):
        ctx.value = ctx.n.value

    def exitMax_value(self, ctx):
        ctx.value = ctx.n.value

    def exitScale(self, ctx):
        ctx.value = ctx.n.value

    def exitOffset(self, ctx):
        ctx.value = ctx.n.value

    def exitText_info(self, ctx):
        ctx.value = ctx.t.value

    def exitStatus_management(self, ctx):
        r = ctx.r.value
        values = [x.value for x in ctx.values] if ctx.values else []
        ctx.value = dict(responseError = r, faultStateSignals = values)

    def exitPublished_signal(self, ctx):
        pass

    def exitFree_text_definition(self, ctx):
        text = ctx.f.value
        ctx.value = text



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
"""

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

    ( C) 2010-2020 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__  = 'Christoph Schueler'
__version__ = '0.1.0'

import argparse
import io
import pathlib
import sys

from pydbc.db.imex import DbcExporter, LdfExporter
import pydbc.db.model as model

def exportFile(pth):
    fnext = pth.suffix[ 1 : ].lower()

    print("Processing '{}'".format(pth))
    #exporter = DbcExporter(pth)
    exporter = LdfExporter(pth, debug = False)

    exporter.run()
    print("OK, done.\n", flush = True)

exportFile(pathlib.Path(sys.argv[1]))


